import unittest
import os
import tempfile
import csv


from ..problem import Problem
from ..algorithm_sweep import SweepAlgorithm
from ..algorithm_genetic import NSGAII
from ..operators import CustomGenerator
from ..results import Results


class TestProblem(Problem):
    """ Describe simple one objective optimization problem. """
    def set(self):
        self.parameters = [{'name': 'x_1', 'initial_value': 10.0, 'bounds': [-10, 20]},
                           {'name': 'x_2', 'initial_value': 10.0, 'bounds': [-10, 20]}]
        self.costs = [{'name': 'F_1'}]

    def evaluate(self, individual):
        result = 0
        for i in individual.vector:
            result += i*i

        return [result]

    def evaluate_constraints(self, individual):
        pass


class TestResults(unittest.TestCase):
    """ Tests simple one objective optimization problem."""

    def setUp(self):
        self.problem = TestProblem()

        generator = CustomGenerator()
        generator.init([[4.0, 2.0], [-1.0, -3.0], [2.0, 4.0]])

        algorithm = SweepAlgorithm(self.problem, generator=generator)
        algorithm.options['max_processes'] = 1
        algorithm.run()

        self.results = Results(self.problem)

    def test_individuals(self):
        self.assertEqual(len(self.problem.individuals), 3)

        self.assertAlmostEqual(self.problem.individuals[0].vector[0], 4.0)
        self.assertAlmostEqual(self.problem.individuals[1].vector[0], -1.0)
        self.assertAlmostEqual(self.problem.individuals[2].vector[0], 2.0)
        self.assertAlmostEqual(self.problem.individuals[0].vector[1], 2.0)
        self.assertAlmostEqual(self.problem.individuals[1].vector[1], -3.0)
        self.assertAlmostEqual(self.problem.individuals[2].vector[1], 4.0)
        self.assertAlmostEqual(self.problem.individuals[0].costs[0], 20.0)
        self.assertAlmostEqual(self.problem.individuals[1].costs[0], 10.0)
        self.assertAlmostEqual(self.problem.individuals[2].costs[0], 20.0)

    def test_parameter_names(self):
        parameter_names = self.results.parameter_names()
        self.assertEqual(parameter_names[0], "x_1")

    def test_goal_names(self):
        goal_names = self.results.goal_names()
        self.assertEqual(goal_names[0], "F_1")

    def test_parameter_number(self):
        self.assertEqual(self.results.parameter_number(), 2)

    def test_goal_number(self):
        self.assertEqual(self.results.goal_number(), 1)

    def test_parameter_index(self):
        self.assertEqual(self.results.parameter_index('x_2'), 1)

    def test_goal_index(self):
        self.assertEqual(self.results.goal_index('F_1'), 0)

    def test_parameters(self):
        parameters = self.results.parameters()
        self.assertEqual(parameters[0], [4.0, 2.0])
        self.assertEqual(parameters[1], [-1.0, -3.0])
        self.assertEqual(parameters[2], [2.0, 4.0])

    def test_costs(self):
        costs = self.results.costs()
        self.assertEqual(len(costs), 3)
        self.assertEqual(costs[0][0], 20.0)
        self.assertEqual(costs[1][0], 10.0)
        self.assertEqual(costs[2][0], 20.0)

    def test_table(self):
        table = self.results.table(transpose=False)
        self.assertEqual(len(table), 3)
        self.assertEqual(table[1][2], 10.0)

        table = self.results.table(transpose=True)
        self.assertEqual(len(table), 3)
        self.assertEqual(table[1][2], 4.0)

    def test_goal_on_index(self):
        goal_on_index = self.results.goal_on_index("F_1")
        self.assertEqual(goal_on_index[0], [0, 1, 2])
        self.assertEqual(goal_on_index[1], [20.0, 10.0, 20.0])

        goal_on_index = self.results.goal_on_index()
        self.assertEqual(goal_on_index[0], [0, 1, 2])
        self.assertEqual(goal_on_index[1], [20.0, 10.0, 20.0])

    def test_parameter_on_index(self):
        parameter_on_index = self.results.parameter_on_index()
        self.assertEqual(parameter_on_index[0], [0, 1, 2])
        self.assertEqual(parameter_on_index[1], [4.0, -1.0, 2.0])
        self.assertEqual(parameter_on_index[2], [2.0, -3.0, 4.0])

        parameter_on_index = self.results.parameter_on_index("x_2")
        self.assertEqual(parameter_on_index[0], [0, 1, 2])
        self.assertEqual(parameter_on_index[1], [2.0, -3.0, 4.0])

    def test_goal_on_parameter(self):
        goal_on_parameter = self.results.goal_on_parameter("x_1", "F_1", sorted=True)
        self.assertEqual(goal_on_parameter[0], [-1.0, 2.0, 4.0])
        self.assertEqual(goal_on_parameter[1], [10.0, 20.0, 20.0])

        goal_on_parameter = self.results.goal_on_parameter("x_1", "F_1")
        self.assertEqual(goal_on_parameter[0], [4.0, -1.0, 2.0])
        self.assertEqual(goal_on_parameter[1], [20.0, 10.0, 20.0])

    def test_parameter_on_goal(self):
        parameter_on_goal = self.results.parameter_on_goal("F_1", "x_1", sorted=True)
        self.assertEqual(parameter_on_goal[0], [10.0, 20.0, 20.0])
        self.assertEqual(parameter_on_goal[1], [-1.0, 2.0, 4.0])

        parameter_on_goal = self.results.parameter_on_goal("F_1", "x_1")
        self.assertEqual(parameter_on_goal[0], [20.0, 10.0, 20.0])
        self.assertEqual(parameter_on_goal[1], [4.0, -1.0, 2.0])

    def test_parameter_on_parameter(self):
        parameter_on_parameter = self.results.parameter_on_parameter("x_1", "x_2", sorted=True)
        self.assertEqual(parameter_on_parameter[0], [-1.0, 2.0, 4.0])
        self.assertEqual(parameter_on_parameter[1], [-3.0, 4.0, 2.0])

        parameter_on_parameter = self.results.parameter_on_parameter("x_1", "x_2")
        self.assertEqual(parameter_on_parameter[0], [4.0, -1.0, 2.0])
        self.assertEqual(parameter_on_parameter[1], [2.0, -3.0, 4.0])

    def test_pareto_front(self):
        problem = TestProblem()

        algorithm = NSGAII(problem)
        algorithm.options['max_processes'] = 10
        algorithm.options['max_population_number'] = 30
        algorithm.options['max_population_size'] = 10
        algorithm.run()

        results = Results(problem)
        pareto_front = results.pareto_front()
        # print(pareto_front)
        # self.assertEqual()
        pass

    def test_export_to_csv(self):
        csv_filename = tempfile.NamedTemporaryFile(mode="w", delete=False, dir=None, suffix=".csv").name
        self.results.export_to_csv(csv_filename)

        with open(csv_filename, 'r', newline='') as f:
            reader = csv.reader(f)
            rows = []
            for row in reader:
                rows.append(row)
            self.assertEqual(rows[0], ['population_id', 'x_1', 'x_2', 'F_1'])
            self.assertEqual(rows[1], ['0', '20.0', '4.0', '2.0'])
            self.assertEqual(rows[2], ['0', '10.0', '-1.0', '-3.0'])
            self.assertEqual(rows[3], ['0', '20.0', '2.0', '4.0'])

        # remove file
        os.remove(csv_filename)

    def test_population(self):
        individuals = self.results.population()
        self.assertEqual(len(individuals), 3)
        self.assertEqual(individuals[1].costs[0], 10.0)




import sqlite3

import os
import json

from individual import Individual


class DummyDataStore:
    def __init__(self):
        pass

    def sync_individual(self, individuals):
        pass

    def sync_all(self):
        pass

    def destroy(self):
        pass


class SqliteDataStore(DummyDataStore):
    sql_main_table = "CREATE TABLE IF NOT EXISTS main (name text NOT NULL, description text NOT NULL);"
    sql_parameters_table = "CREATE TABLE IF NOT EXISTS parameters (name text PRIMARY KEY, parameter json not null);"
    sql_costs_table = "CREATE TABLE IF NOT EXISTS costs (name text PRIMARY KEY, cost json not null);"
    sql_individuals_table = "CREATE TABLE IF NOT EXISTS individuals (id int PRIMARY KEY, individual json not null);"

    sql_main_insert = "INSERT INTO main(name, description) VALUES (?,?)"
    sql_parameters_insert = "INSERT INTO parameters(name, parameter) VALUES (?,?)"
    sql_costs_insert = "INSERT INTO costs(name, cost) VALUES (?,?)"

    sql_individuals_upsert = "INSERT INTO individuals (id, individual) VALUES(?,?) ON CONFLICT(id) DO UPDATE SET individual=excluded.individual;"

    sql_main_select = "SELECT * FROM main;"
    sql_parameters_select = "SELECT * FROM parameters;"
    sql_costs_select = "SELECT * FROM costs;"
    sql_individuals_select = "SELECT * FROM individuals;"
    # SELECT json_extract(individual, '$.costs[0]') as cost FROM individuals where cost>10;

    def __init__(self, problem, database_name, mode="write", thread_safe=True):
        self.problem = problem
        self.database_name = database_name
        self.mode = mode
        self.thread_safe = thread_safe
        # cache
        self._conn = None

        if not self.database_name:
            raise RuntimeError("SqliteDataStoreCacheThread: database name is empty.")

        if self.mode == "write":
            if os.path.exists(self.database_name):
                statinfo = os.stat(self.database_name)
                if statinfo.st_size == 0:
                    self._create_structure()
                else:
                    self.read_from_datastore()
            else:
                self._create_structure()
        elif self.mode == "rewrite":
            if os.path.exists(database_name):
                os.remove(database_name)
            self._create_structure()
        elif self.mode == "read":
            self.read_from_datastore()

    def conn(self):
        if self.thread_safe:
            try:
                if self.mode == "write":
                    conn = sqlite3.connect(self.database_name, isolation_level='Exclusive')
                    c = conn.cursor()
                    c.execute('PRAGMA synchronous = 0')
                    c.execute('PRAGMA journal_mode = ON')
                    conn.commit()
                else:
                    conn = sqlite3.connect(self.database_name)
            except sqlite3.Error as e:
                print(e)

            return conn
        else:
            if self._conn is None:
                try:
                    if self.mode == "write":
                        self._conn = sqlite3.connect(self.database_name, isolation_level='Exclusive')
                        c = self._conn.cursor()
                        c.execute('PRAGMA synchronous = 0')
                        c.execute('PRAGMA journal_mode = OFF')
                        self._conn.commit()
                    else:
                        self._conn = sqlite3.connect(self.database_name)
                except sqlite3.Error as e:
                    print(e)

            return self._conn

    def _create_structure(self):
        conn = self.conn()
        c = conn.cursor()

        # structure
        c.execute(self.sql_main_table)
        c.execute(self.sql_costs_table)
        c.execute(self.sql_parameters_table)
        c.execute(self.sql_individuals_table)
        conn.commit()

        # data
        c.execute(self.sql_main_insert, [self.problem.name, self.problem.description])
        for parameter in self.problem.parameters:
            c.execute(self.sql_parameters_insert, [parameter["name"], json.dumps(parameter)])
        for cost in self.problem.costs:
            c.execute(self.sql_costs_insert, [cost["name"], json.dumps(cost)])
        conn.commit()

    def read_from_datastore(self):
        conn = self.conn()
        c = conn.cursor()

        c.execute(self.sql_main_select)
        rows = c.fetchall()

        # main
        self.problem.name = rows[0][0]
        self.problem.description = rows[0][1]

        # parameters
        self.problem.parameters.clear()
        c.execute(self.sql_parameters_select)
        rows = c.fetchall()
        for row in rows:
            parameter = json.loads(row[1])
            self.problem.parameters.append(parameter)

        # costs
        self.problem.costs.clear()
        c.execute(self.sql_costs_select)
        rows = c.fetchall()
        for row in rows:
            cost = json.loads(row[1])
            self.problem.costs.append(cost)

        # individuals
        self.problem.individuals.clear()
        c.execute(self.sql_individuals_select)
        rows = c.fetchall()
        for row in rows:
            individual = Individual.from_dict(json.loads(row[1]))
            self.problem.individuals.append(individual)

    def destroy(self):
        if self._conn is not None:
            self._conn.close()

    def sync_individual(self, individual):
        if self.mode == "write" or self.mode == "rewrite":
            conn = self.conn()
            c = conn.cursor()

            # data
            try:
                c.execute(self.sql_individuals_upsert, [individual.id, json.dumps(individual.to_dict())])
                conn.commit()
            except sqlite3.OperationalError as e:
                # try again
                self.sync_individual(individual)

    def sync_all(self):
        if self.mode == "write" or self.mode == "rewrite":
            conn = self.conn()
            c = conn.cursor()

            for individual in self.problem.individuals:
                c.execute(self.sql_individuals_upsert, [individual.id, json.dumps(individual.to_dict())])

            conn.commit()



import math
import unittest

from ..individual import Individual
from ..benchmark_functions import Booth
from ..surrogate import SurrogateModelEval


class TestSurrogateEval(unittest.TestCase):
    def test_eval(self):
        problem = Booth()
        problem.surrogate = SurrogateModelEval(problem)

        x_ref = Individual([2.5, 1.5])
        # eval reference
        value_problem = problem.evaluate(x_ref)[0]
        # eval surrogate
        value_surrogate = problem.surrogate.predict(x_ref)[0]

        problem.logger.info(
            "{}: surrogate.value: evaluation = {}, prediction = {}, difference = {}".format(problem.name,
                                                                                            value_problem,
                                                                                            value_surrogate,
                                                                                            math.fabs(value_problem -
                                                                                                      value_surrogate)))
        self.assertLess(math.fabs(value_problem - value_surrogate), 1e-8)



import numpy as np
from scipy.linalg import toeplitz, hankel, cholesky
from scipy import spatial
from scipy import stats
from numpy import ma
import re
import itertools

"""
This code was originally published by the following individuals for use with
Scilab:
    Copyright (C) 2012 - 2013 - Michael Baudin
    Copyright (C) 2012 - Maria Christopoulou
    Copyright (C) 2010 - 2011 - INRIA - Michael Baudin
    Copyright (C) 2009 - Yann Collette
    Copyright (C) 2009 - CEA - Jean-Marc Martinez

    website: forge.scilab.org/index.php/p/scidoe/sourcetree/master/macros
Much thanks goes to these individuals. It has been converted to Python by 
Abraham Lee.

https://github.com/tirthajyoti/Design-of-experiment-Python
https://pythonhosted.org/pyDOE/

GSD Copyright (C) 2018 - Rickard Sjoegren
https://github.com/clicumu/pyDOE2/blob/master/pyDOE2
"""


# Function for constructing a DataFrame from a numpy array generated by PyDOE function and individual lists
def construct_df(x, factor_lists):
    out = []
    for col in x:
        row = []
        for index in range(len(col)):
            row.append(factor_lists[index][int(col[index])])
        out.append(row)

    return out


# Function for constructing a DataFrame from a matrix with floating point numbers between -1 and +1
def construct_df_from_matrix(x, factor_array):
    assert 0

    """
    row_num = x.shape[0]  # Number of rows in the matrix x
    col_num = x.shape[1]  # Number of columns in the matrix x

    empty = np.zeros((row_num, col_num))

    def simple_substitution(idx, factor_list):
        if idx == -1:
            return factor_list[0]
        elif idx == 0:
            return factor_list[1]
        elif idx == 1:
            return factor_list[2]
        else:
            alpha = np.abs(factor_list[2] - factor_list[0]) / 2
            if idx < 0:
                beta = np.abs(idx) - 1
                return factor_list[0] - (beta * alpha)
            else:
                beta = idx - 1
                return factor_list[2] + (beta * alpha)

    for i in range(row_num):
        for j in range(col_num):
            empty[i, j] = simple_substitution(x[i, j], factor_array[j])

    return pd.DataFrame(data=empty)
    """

# Function for constructing a DataFrame from a matrix with floating point numbers between 0 and 1
def construct_df_from_random_matrix(x, factor_lists):
    out = []
    for w in x:
        row = []
        for index in range(len(w)):
            row.append(factor_lists[index][0] + (w[index] * np.fabs(factor_lists[index][1] - factor_lists[index][0])))

        out.append(row)

    return out


# Function for building Box-Behnken designs from a dictionary of process variables
def build_box_behnken(factor_level_ranges):
    """
    Builds a Box-Behnken design dataframe from a dictionary of factor/level ranges.
    Note 3 levels of factors are necessary. If not given, the function will automatically create 3 levels by linear mid-section method.
    Example of the dictionary:
    {'Pressure':[50, 70],'Temperature':[290, 350],'Flow rate':[0.9, 1.1]}

	In statistics, Box–Behnken designs are experimental designs for response surface methodology, devised by George E. P. Box and Donald Behnken in 1960, to achieve the following goals:
		* Each factor, or independent variable, is placed at one of three equally spaced values, usually coded as −1, 0, +1. (At least three levels are needed for the following goal.)
		* The design should be sufficient to fit a quadratic model, that is, one containing squared terms, products of two factors, linear terms and an intercept.
		* The ratio of the number of experimental points to the number of coefficients in the quadratic model should be reasonable (in fact, their designs kept it in the range of 1.5 to 2.6).*estimation variance should more or less depend only on the distance from the centre (this is achieved exactly for the designs with 4 and 7 factors), and should not vary too much inside the smallest (hyper)cube containing the experimental points.
	"""

    # creating a mid-point by averaging them
    for key in factor_level_ranges:
        if len(factor_level_ranges[key]) == 2:
            factor_level_ranges[key].append((factor_level_ranges[key][0] + factor_level_ranges[key][1]) / 2)
            factor_level_ranges[key].sort()


    factor_count = len(factor_level_ranges)
    factor_lists = []

    for key in factor_level_ranges:
        factor_lists.append(factor_level_ranges[key])

    x = bbdesign(factor_count, center=1)
    x = x + 1  # Adjusting the index up by 1

    df = construct_df(x, factor_lists)

    return df


def bbdesign(n, center=None):
    """
    Create a Box-Behnken design

    Parameters
    ----------
    n : int
        The number of factors in the design

    Optional
    --------
    center : int
        The number of center points to include (default = 1).

    Returns
    -------
    mat : 2d-array
        The design matrix

    Example
    -------
    ::

        >>> bbdesign(3)
        array([[-1., -1.,  0.],
               [ 1., -1.,  0.],
               [-1.,  1.,  0.],
               [ 1.,  1.,  0.],
               [-1.,  0., -1.],
               [ 1.,  0., -1.],
               [-1.,  0.,  1.],
               [ 1.,  0.,  1.],
               [ 0., -1., -1.],
               [ 0.,  1., -1.],
               [ 0., -1.,  1.],
               [ 0.,  1.,  1.],
               [ 0.,  0.,  0.],
               [ 0.,  0.,  0.],
               [ 0.,  0.,  0.]])

    """
    assert n >= 3, 'Number of variables must be at least 3'

    # First, compute a factorial DOE with 2 parameters
    H_fact = ff2n(2)
    # Now we populate the real DOE with this DOE

    # We made a factorial design on each pair of dimensions
    # - So, we created a factorial design with two factors
    # - Make two loops
    Index = 0
    nb_lines = int((0.5 * n * (n - 1)) * H_fact.shape[0])
    H = repeat_center(n, nb_lines)

    for i in range(n - 1):
        for j in range(i + 1, n):
            Index = Index + 1
            H[max([0, (Index - 1) * H_fact.shape[0]]):Index * H_fact.shape[0], i] = H_fact[:, 0]
            H[max([0, (Index - 1) * H_fact.shape[0]]):Index * H_fact.shape[0], j] = H_fact[:, 1]

    if center is None:
        if n <= 16:
            points = [0, 0, 0, 3, 3, 6, 6, 6, 8, 9, 10, 12, 12, 13, 14, 15, 16]
            center = points[n]
        else:
            center = n

    H = np.c_[H.T, repeat_center(n, center).T].T

    return H


def pbdesign(n):
    """
    Generate a Plackett-Burman design

    Parameter
    ---------
    n : int
        The number of factors to create a matrix for.

    Returns
    -------
    H : 2d-array
        An orthogonal design matrix with n columns, one for each factor, and
        the number of rows being the next multiple of 4 higher than n (e.g.,
        for 1-3 factors there are 4 rows, for 4-7 factors there are 8 rows,
        etc.)

    Example
    -------

    A 3-factor design::

        >>> pbdesign(3)
        array([[-1., -1.,  1.],
               [ 1., -1., -1.],
               [-1.,  1., -1.],
               [ 1.,  1.,  1.]])

    A 5-factor design::

        >>> pbdesign(5)
        array([[-1., -1.,  1., -1.,  1.],
               [ 1., -1., -1., -1., -1.],
               [-1.,  1., -1., -1.,  1.],
               [ 1.,  1.,  1., -1., -1.],
               [-1., -1.,  1.,  1., -1.],
               [ 1., -1., -1.,  1.,  1.],
               [-1.,  1., -1.,  1., -1.],
               [ 1.,  1.,  1.,  1.,  1.]])

    """
    assert n > 0, 'Number of factors must be a positive integer'
    keep = int(n)
    n = 4 * (int(n / 4) + 1)  # calculate the correct number of rows (multiple of 4)
    f, e = np.frexp([n, n / 12., n / 20.])
    k = [idx for idx, val in enumerate(np.logical_and(f == 0.5, e > 0)) if val]

    assert isinstance(n, int) and k != [], 'Invalid inputs. n must be a multiple of 4.'

    k = k[0]
    e = e[k] - 1

    if k == 0:  # N = 1*2**e
        H = np.ones((1, 1))
    elif k == 1:  # N = 12*2**e
        H = np.vstack((np.ones((1, 12)), np.hstack((np.ones((11, 1)),
                                                    toeplitz([-1, -1, 1, -1, -1, -1, 1, 1, 1, -1, 1],
                                                             [-1, 1, -1, 1, 1, 1, -1, -1, -1, 1, -1])))))
    elif k == 2:  # N = 20*2**e
        H = np.vstack((np.ones((1, 20)), np.hstack((np.ones((19, 1)),
                                                    hankel(
                                                        [-1, -1, 1, 1, -1, -1, -1, -1, 1, -1, 1, -1, 1, 1, 1, 1, -1, -1, 1],
                                                        [1, -1, -1, 1, 1, -1, -1, -1, -1, 1, -1, 1, -1, 1, 1, 1, 1, -1, -1])
                                                    ))))

    # Kronecker product construction
    for i in range(e):
        H = np.vstack((np.hstack((H, H)), np.hstack((H, -H))))

    # Reduce the size of the matrix as needed
    H = H[:, 1:(keep + 1)]

    return np.flipud(H)


def repeat_center(n, repeat):
    """
    Create the center-point portion of a design matrix

    Parameters
    ----------
    n : int
        The number of factors in the original design
    repeat : int
        The number of center points to repeat

    Returns
    -------
    mat : 2d-array
        The center-point portion of a design matrix (elements all zero).

    Example
    -------
    ::

        >>> repeat_center(3, 2)
        array([[ 0.,  0.,  0.],
               [ 0.,  0.,  0.]])

    """
    return np.zeros((repeat, n))


def fullfact(levels):
    """
    Create a general full-factorial design

    Parameters
    ----------
    levels : array-like
        An array of integers that indicate the number of levels of each input
        design factor.

    Returns
    -------
    mat : 2d-array
        The design matrix with coded levels 0 to k-1 for a k-level factor

    Example
    -------
    ::

        >>> fullfact([2, 4, 3])
        array([[ 0.,  0.,  0.],
               [ 1.,  0.,  0.],
               [ 0.,  1.,  0.],
               [ 1.,  1.,  0.],
               [ 0.,  2.,  0.],
               [ 1.,  2.,  0.],
               [ 0.,  3.,  0.],
               [ 1.,  3.,  0.],
               [ 0.,  0.,  1.],
               [ 1.,  0.,  1.],
               [ 0.,  1.,  1.],
               [ 1.,  1.,  1.],
               [ 0.,  2.,  1.],
               [ 1.,  2.,  1.],
               [ 0.,  3.,  1.],
               [ 1.,  3.,  1.],
               [ 0.,  0.,  2.],
               [ 1.,  0.,  2.],
               [ 0.,  1.,  2.],
               [ 1.,  1.,  2.],
               [ 0.,  2.,  2.],
               [ 1.,  2.,  2.],
               [ 0.,  3.,  2.],
               [ 1.,  3.,  2.]])

    """
    n = len(levels)  # number of factors
    nb_lines = np.prod(levels)  # number of trial conditions
    H = np.zeros((nb_lines, n))

    level_repeat = 1
    range_repeat = np.prod(levels)
    for i in range(n):
        range_repeat //= levels[i]
        lvl = []
        for j in range(levels[i]):
            lvl += [j] * level_repeat
        rng = lvl * range_repeat
        level_repeat *= levels[i]
        H[:, i] = rng

    return H


################################################################################

def fracfact(gen):
    """
    Create a 2-level fractional-factorial design with a generator string.

    Parameters
    ----------
    gen : str
        A string, consisting of lowercase, uppercase letters or operators "-"
        and "+", indicating the factors of the experiment

    Returns
    -------
    H : 2d-array
        A m-by-n matrix, the fractional factorial design. m is 2^k, where k
        is the number of letters in ``gen``, and n is the total number of
        entries in ``gen``.

    Notes
    -----
    In ``gen`` we define the main factors of the experiment and the factors
    whose levels are the products of the main factors. For example, if

        gen = "a b ab"

    then "a" and "b" are the main factors, while the 3rd factor is the product
    of the first two. If we input uppercase letters in ``gen``, we get the same
    result. We can also use the operators "+" and "-" in ``gen``.

    For example, if

        gen = "a b -ab"

    then the 3rd factor is the opposite of the product of "a" and "b".

    The output matrix includes the two level full factorial design, built by
    the main factors of ``gen``, and the products of the main factors. The
    columns of ``H`` follow the sequence of ``gen``.

    For example, if

        gen = "a b ab c"

    then columns H[:, 0], H[:, 1], and H[:, 3] include the two level full
    factorial design and H[:, 2] includes the products of the main factors.

    Examples
    --------
    ::

        >>> fracfact("a b ab")
        array([[-1., -1.,  1.],
               [ 1., -1., -1.],
               [-1.,  1., -1.],
               [ 1.,  1.,  1.]])

        >>> fracfact("A B AB")
        array([[-1., -1.,  1.],
               [ 1., -1., -1.],
               [-1.,  1., -1.],
               [ 1.,  1.,  1.]])

        >>> fracfact("a b -ab c +abc")
        array([[-1., -1., -1., -1., -1.],
               [ 1., -1.,  1., -1.,  1.],
               [-1.,  1.,  1., -1.,  1.],
               [ 1.,  1., -1., -1., -1.],
               [-1., -1., -1.,  1.,  1.],
               [ 1., -1.,  1.,  1., -1.],
               [-1.,  1.,  1.,  1., -1.],
               [ 1.,  1., -1.,  1.,  1.]])

    """
    # Recognize letters and combinations
    A = [item for item in re.split('\-?\s?\+?', gen) if item]  # remove empty strings
    C = [len(item) for item in A]

    # Indices of single letters (main factors)
    I = [i for i, item in enumerate(C) if item == 1]

    # Indices of letter combinations (we need them to fill out H2 properly).
    J = [i for i, item in enumerate(C) if item != 1]

    # Check if there are "-" or "+" operators in gen
    U = [item for item in gen.split(' ') if item]  # remove empty strings

    # If R1 is either None or not, the result is not changed, since it is a
    # multiplication of 1.
    R1 = _grep(U, '+')
    R2 = _grep(U, '-')

    # Fill in design with two level factorial design
    H1 = ff2n(len(I))
    H = np.zeros((H1.shape[0], len(C)))
    H[:, I] = H1

    # Recognize combinations and fill in the rest of matrix H2 with the proper
    # products
    for k in J:
        # For lowercase letters
        xx = np.array([ord(c) for c in A[k]]) - 97

        # For uppercase letters
        if np.any(xx < 0):
            xx = np.array([ord(c) for c in A[k]]) - 65

        H[:, k] = np.prod(H1[:, xx], axis=1)

    # Update design if gen includes "-" operator
    if R2:
        H[:, R2] *= -1

    # Return the fractional factorial design
    return H


def _grep(haystack, needle):
    try:
        haystack[0]
    except (TypeError, AttributeError):
        return [0] if needle in haystack else []
    else:
        locs = []
        for idx, item in enumerate(haystack):
            if needle in item:
                locs += [idx]
        return locs


# Function for building full factorial DataFrame from a dictionary of process variables
# ======================================================================================

def build_full_fact(factor_level_ranges):
    """
    Builds a full factorial design dataframe from a dictionary of factor/level ranges
    Example of the process variable dictionary:
    {'Pressure':[50,60,70],'Temperature':[290, 320, 350],'Flow rate':[0.9,1.0]}
    """

    factor_lvl_count = []
    factor_lists = []

    for key in factor_level_ranges:
        factor_lvl_count.append(len(factor_level_ranges[key]))
        factor_lists.append(factor_level_ranges[key])

    x = fullfact(factor_lvl_count)
    df = construct_df(x, factor_lists)

    return df


# ================================================================================================================================================================
# Function for building 2-level fractional factorial DataFrame from a dictionary of process variables and a generator string representing the confounded variable
# ================================================================================================================================================================

def build_frac_fact(factor_level_ranges, gen_string):
    """
    Builds a full factorial design dataframe from a dictionary of factor/level ranges.
    Only min and max values of the range are required.
	Example of the dictionary:
    {'Pressure':[50,70],'Temperature':[290, 350],'Flow rate':[0.9,1.0]}

	This function requires a little more knowledge of how the confounding will be allowed.
	This means that some factor effects get muddled with other interaction effects, so it’s harder to distinguish between them).

	Let’s assume that we just can’t afford (for whatever reason) the number of runs in a full-factorial design. We can systematically decide on a fraction of the full-factorial by allowing some of the factor main effects to be confounded with other factor interaction effects.
	This is done by defining an alias structure that defines, symbolically, these interactions. These alias structures are written like “C = AB” or “I = ABC”, or “AB = CD”, etc.
	These define how one column is related to the others.

	EXAMPLE
    ------------
    For example, the alias “C = AB” or “I = ABC” indicate that there are three factors (A, B, and C) and that the main effect of factor C is confounded with the interaction effect of the product AB, and by extension, A is confounded with BC and B is confounded with AC.
	A full- factorial design with these three factors results in a design matrix with 8 runs, but we will assume that we can only afford 4 of those runs.
	To create this fractional design, we need a matrix with three columns, one for A, B, and C, only now where the levels in the C column is created by the product of the A and B columns.
    """

    factor_count = len(factor_level_ranges)
    factor_lists = []

    # Assigning the end point to the high level.
    for key in factor_level_ranges:
        if len(factor_level_ranges[key]) != 2:
            factor_level_ranges[key][1] = factor_level_ranges[key][-1]
            factor_level_ranges[key] = factor_level_ranges[key][:2]

    if factor_count != len(gen_string.split(' ')):
        print("Length of the generator string for the fractional factorial build does not match the length of the process variables dictionary")
        return None

    for key in factor_level_ranges:
        factor_lists.append(factor_level_ranges[key])

    x = fracfact(gen_string)

    def index_change(x):
        if x == -1:
            return 0
        else:
            return x

    vfunc = np.vectorize(index_change)
    x = vfunc(x)

    df = construct_df(x, factor_lists)

    return df


# =====================================================================================
# Function for building Plackett-Burman designs from a dictionary of process variables
# =====================================================================================

def build_plackett_burman(factor_level_ranges):
    """
    Builds a Plackett-Burman dataframe from a dictionary of factor/level ranges.
    Only min and max values of the range are required.
    Example of the dictionary:
    {'Pressure':[50,70],'Temperature':[290, 350],'Flow rate':[0.9,1.0]}

	Plackett–Burman designs are experimental designs presented in 1946 by Robin L. Plackett and J. P. Burman while working in the British Ministry of Supply.(Their goal was to find experimental designs for investigating the dependence of some measured quantity on a number of independent variables (factors), each taking L levels, in such a way as to minimize the variance of the estimates of these dependencies using a limited number of experiments.

    Interactions between the factors were considered negligible. The solution to this problem is to find an experimental design where each combination of levels for any pair of factors appears the same number of times, throughout all the experimental runs (refer to table).
	A complete factorial design would satisfy this criterion, but the idea was to find smaller designs.

	These designs are unique in that the number of trial conditions (rows) expands by multiples of four (e.g. 4, 8, 12, etc.).
	The max number of columns allowed before a design increases the number of rows is always one less than the next higher multiple of four.
    """

    # Assigning the end point to the high level.
    for key in factor_level_ranges:
        if len(factor_level_ranges[key]) != 2:
            factor_level_ranges[key][1] = factor_level_ranges[key][-1]
            factor_level_ranges[key] = factor_level_ranges[key][:2]

    factor_count = len(factor_level_ranges)
    factor_lists = []

    for key in factor_level_ranges:
        factor_lists.append(factor_level_ranges[key])

    x = pbdesign(factor_count)

    def index_change(x):
        if x == -1:
            return 0
        else:
            return x

    vfunc = np.vectorize(index_change)
    x = vfunc(x)

    df = construct_df(x, factor_lists)
    

    return df


# ===================================================================================
# Function for building Box-Behnken designs from a dictionary of process variables
# ===================================================================================

def build_box_behnken(factor_level_ranges):
    """
    Builds a Box-Behnken design dataframe from a dictionary of factor/level ranges.
    Note 3 levels of factors are necessary. If not given, the function will automatically create 3 levels by linear mid-section method.
    Example of the dictionary:
    {'Pressure':[50,60,70],'Temperature':[290, 320, 350],'Flow rate':[0.9,1.0,1.1]}

	In statistics, Box–Behnken designs are experimental designs for response surface methodology, devised by George E. P. Box and Donald Behnken in 1960, to achieve the following goals:
		* Each factor, or independent variable, is placed at one of three equally spaced values, usually coded as −1, 0, +1. (At least three levels are needed for the following goal.)
		* The design should be sufficient to fit a quadratic model, that is, one containing squared terms, products of two factors, linear terms and an intercept.
		* The ratio of the number of experimental points to the number of coefficients in the quadratic model should be reasonable (in fact, their designs kept it in the range of 1.5 to 2.6).*estimation variance should more or less depend only on the distance from the centre (this is achieved exactly for the designs with 4 and 7 factors), and should not vary too much inside the smallest (hyper)cube containing the experimental points.
	"""

    # Creating a mid-point by averaging them
    for key in factor_level_ranges:
        if len(factor_level_ranges[key]) == 2:
            factor_level_ranges[key].append((factor_level_ranges[key][0] + factor_level_ranges[key][1]) / 2)
            factor_level_ranges[key].sort()

    factor_count = len(factor_level_ranges)
    factor_lists = []

    for key in factor_level_ranges:
        factor_lists.append(factor_level_ranges[key])

    x = bbdesign(factor_count, center=1)
    x = x + 1  # Adjusting the index up by 1

    df = construct_df(x, factor_lists)

    return df


# Function for building central-composite (Box-Wilson) designs from a dictionary of process variables
def build_central_composite(factor_level_ranges, center=(2, 2), alpha='o', face='ccc'):
    """
    Builds a central-composite design dataframe from a dictionary of factor/level ranges.
    Only min and max values of the range are required.
    Example of the dictionary:
    {'Pressure':[50,70],'Temperature':[290, 350],'Flow rate':[0.9,1.0]}

	In statistics, a central composite design is an experimental design, useful in response surface methodology, for building a second order (quadratic) model for the response variable without needing to use a complete three-level factorial experiment.
	The design consists of three distinct sets of experimental runs:
		* A factorial (perhaps fractional) design in the factors studied, each having two levels;
		* A set of center points, experimental runs whose values of each factor are the medians of the values used in the factorial portion. This point is often replicated in order to improve the precision of the experiment;
		* A set of axial points, experimental runs identical to the centre points except for one factor, which will take on values both below and above the median of the two factorial levels, and typically both outside their range. All factors are varied in this way.
    """
    # Assigning the end point to the high level.
    for key in factor_level_ranges:
        if len(factor_level_ranges[key]) != 2:
            factor_level_ranges[key][1] = factor_level_ranges[key][-1]
            factor_level_ranges[key] = factor_level_ranges[key][:2]

    # Creates the mid-points by averaging the low and high levels
    for key in factor_level_ranges:
        if len(factor_level_ranges[key]) == 2:
            factor_level_ranges[key].append((factor_level_ranges[key][0] + factor_level_ranges[key][1]) / 2)
            factor_level_ranges[key].sort()

    factor_count = len(factor_level_ranges)
    factor_lists = []

    for key in factor_level_ranges:
        factor_lists.append(factor_level_ranges[key])

    x = ccdesign(factor_count, center=center, alpha=alpha, face=face)
    factor_lists = np.array(factor_lists)

    # rescale x from (-1, 1) to (0, 1)

    df = construct_df_from_matrix(x, factor_lists)
    
    return df


def build_lhs(factor_level_ranges, num_samples=None):
    """
    Builds a Latin Hypercube design dataframe from a dictionary of factor/level ranges.
    Only min and max values of the range are required.
    Example of the dictionary:
    {'Pressure':[50,70],'Temperature':[290, 350],'Flow rate':[0.9,1.0]}
    num_samples: Number of samples to be generated

	Latin hypercube sampling (LHS) is a form of stratified sampling that can be applied to multiple variables. The method commonly used to reduce the number or runs necessary for a Monte Carlo simulation to achieve a reasonably accurate random distribution. LHS can be incorporated into an existing Monte Carlo model fairly easily, and work with variables following any analytical probability distribution.
    """

    # Assigning the end point to the high level.
    for key in factor_level_ranges:
        if len(factor_level_ranges[key]) != 2:
            factor_level_ranges[key][1] = factor_level_ranges[key][-1]
            factor_level_ranges[key] = factor_level_ranges[key][:2]

    factor_count = len(factor_level_ranges)
    factor_lists = []

    if num_samples == None:
        num_samples = factor_count

    for key in factor_level_ranges:
        factor_lists.append(factor_level_ranges[key])

    x = lhs(n=factor_count, samples=num_samples)
    factor_lists = np.array(factor_lists)

    df = construct_df_from_random_matrix(x, factor_lists)

    return df


def lhs(n, samples=None, criterion=None, iterations=None, random_state=None, correlation_matrix=None):
    """
    Generate a latin-hypercube design
    Parameters
    ----------
    n : int
        The number of factors to generate samples for
    Optional
    --------
    samples : int
        The number of samples to generate for each factor (Default: n)
    criterion : str
        Allowable values are "center" or "c", "maximin" or "m",
        "centermaximin" or "cm", and "correlation" or "corr". If no value
        given, the design is simply randomized.
    iterations : int
        The number of iterations in the maximin and correlations algorithms
        (Default: 5).
    randomstate : np.random.RandomState, int
         Random state (or seed-number) which controls the seed and random draws
    correlation_matrix : ndarray
         Enforce correlation between factors (only used in lhsmu)
    Returns
    -------
    H : 2d-array
        An n-by-samples design matrix that has been normalized so factor values
        are uniformly spaced between zero and one.
    Example
    -------
    A 3-factor design (defaults to 3 samples)::
        >>> lhs(3, random_state=42)
        array([[ 0.12484671,  0.95539205,  0.24399798],
               [ 0.53288616,  0.38533955,  0.86703834],
               [ 0.68602787,  0.31690477,  0.38533151]])
    A 4-factor design with 6 samples::
        >>> lhs(4, samples=6, random_state=42)
        array([[ 0.06242335,  0.19266575,  0.88202411,  0.89439364],
               [ 0.19266977,  0.53538985,  0.53030416,  0.49498498],
               [ 0.71737371,  0.75412607,  0.17634727,  0.71520486],
               [ 0.63874044,  0.85658231,  0.33676408,  0.31102936],
               [ 0.43351917,  0.45134543,  0.12199899,  0.53056742],
               [ 0.93530882,  0.15845238,  0.7386575 ,  0.09977641]])
    A 2-factor design with 5 centered samples::
        >>> lhs(2, samples=5, criterion='center', random_state=42)
        array([[ 0.1,  0.9],
               [ 0.5,  0.5],
               [ 0.7,  0.1],
               [ 0.3,  0.7],
               [ 0.9,  0.3]])
    A 3-factor design with 4 samples where the minimum distance between
    all samples has been maximized::
        >>> lhs(3, samples=4, criterion='maximin', random_state=42)
        array([[ 0.69754389,  0.2997106 ,  0.96250964],
               [ 0.10585037,  0.09872038,  0.73157522],
               [ 0.25351996,  0.65148999,  0.07337204],
               [ 0.91276926,  0.97873992,  0.42783549]])
    A 4-factor design with 5 samples where the samples are as uncorrelated
    as possible (within 10 iterations)::
        >>> lhs(4, samples=5, criterion='correlation', iterations=10, random_state=42)
        array([[ 0.72088348,  0.05121366,  0.97609357,  0.92487081],
               [ 0.49507404,  0.51265511,  0.00808672,  0.37915272],
               [ 0.22217816,  0.2878673 ,  0.24034384,  0.42786629],
               [ 0.91977309,  0.93895699,  0.64061224,  0.14213258],
               [ 0.04719698,  0.70796822,  0.53910322,  0.78857071]])
    """
    H = None

    if random_state is None:
        random_state = np.random.RandomState()
    elif not isinstance(random_state, np.random.RandomState):
        random_state = np.random.RandomState(random_state)

    if samples is None:
        samples = n

    if criterion is not None:
        if not criterion.lower() in ('center', 'c', 'maximin', 'm',
                                     'centermaximin', 'cm', 'correlation',
                                     'corr', 'lhsmu'):
            raise ValueError('Invalid value for "criterion": {}'.format(criterion))

    else:
        H = _lhsclassic(n, samples, random_state)

    if criterion is None:
        criterion = 'center'
    if iterations is None:
        iterations = 5

    if H is None:
        if criterion.lower() in ('center', 'c'):
            H = _lhscentered(n, samples, random_state)
        elif criterion.lower() in ('maximin', 'm'):
            H = _lhsmaximin(n, samples, iterations, 'maximin', random_state)
        elif criterion.lower() in ('centermaximin', 'cm'):
            H = _lhsmaximin(n, samples, iterations, 'centermaximin', random_state)
        elif criterion.lower() in ('correlation', 'corr'):
            H = _lhscorrelate(n, samples, iterations, random_state)
        elif criterion.lower() in ('lhsmu'):
            # as specified by the paper. M is set to 5
            H = _lhsmu(n, samples, correlation_matrix, random_state, M=5)

    return H


################################################################################

def _lhsclassic(n, samples, randomstate):
    # Generate the intervals
    cut = np.linspace(0, 1, samples + 1)

    # Fill points uniformly in each interval
    u = randomstate.rand(samples, n)
    a = cut[:samples]
    b = cut[1:samples + 1]
    rdpoints = np.zeros_like(u)
    for j in range(n):
        rdpoints[:, j] = u[:, j] * (b - a) + a

    # Make the random pairings
    H = np.zeros_like(rdpoints)
    for j in range(n):
        order = randomstate.permutation(range(samples))
        H[:, j] = rdpoints[order, j]

    return H


################################################################################

def _lhscentered(n, samples, randomstate):
    # Generate the intervals
    cut = np.linspace(0, 1, samples + 1)

    # Fill points uniformly in each interval
    u = randomstate.rand(samples, n)
    a = cut[:samples]
    b = cut[1:samples + 1]
    _center = (a + b) / 2

    # Make the random pairings
    H = np.zeros_like(u)
    for j in range(n):
        H[:, j] = randomstate.permutation(_center)

    return H


################################################################################

def _lhsmaximin(n, samples, iterations, lhstype, randomstate):
    maxdist = 0

    # Maximize the minimum distance between points
    for i in range(iterations):
        if lhstype == 'maximin':
            Hcandidate = _lhsclassic(n, samples, randomstate)
        else:
            Hcandidate = _lhscentered(n, samples, randomstate)

        d = spatial.distance.pdist(Hcandidate, 'euclidean')
        if maxdist < np.min(d):
            maxdist = np.min(d)
            H = Hcandidate.copy()

    return H


################################################################################

def _lhscorrelate(n, samples, iterations, randomstate):
    mincorr = np.inf

    # Minimize the components correlation coefficients
    for i in range(iterations):
        # Generate a random LHS
        Hcandidate = _lhsclassic(n, samples, randomstate)
        R = np.corrcoef(Hcandidate.T)
        if np.max(np.abs(R[R != 1])) < mincorr:
            mincorr = np.max(np.abs(R - np.eye(R.shape[0])))
            H = Hcandidate.copy()

    return H


################################################################################

def _lhsmu(N, samples=None, corr=None, random_state=None, M=5):
    if random_state is None:
        random_state = np.random.RandomState()
    elif not isinstance(random_state, np.random.RandomState):
        random_state = np.random.RandomState(random_state)

    if samples is None:
        samples = N

    I = M * samples

    rdpoints = random_state.uniform(size=(I, N))

    dist = spatial.distance.cdist(rdpoints, rdpoints, metric='euclidean')
    D_ij = ma.masked_array(dist, mask=np.identity(I))

    index_rm = np.zeros(I - samples, dtype=int)
    i = 0
    while i < I - samples:
        order = ma.sort(D_ij, axis=1)

        avg_dist = ma.mean(order[:, 0:2], axis=1)
        min_l = ma.argmin(avg_dist)

        D_ij[min_l, :] = ma.masked
        D_ij[:, min_l] = ma.masked

        index_rm[i] = min_l
        i += 1

    rdpoints = np.delete(rdpoints, index_rm, axis=0)

    if (corr is not None):
        # check if covariance matrix is valid
        assert type(corr) == np.ndarray
        assert corr.ndim == 2
        assert corr.shape[0] == corr.shape[1]
        assert corr.shape[0] == N

        norm_u = stats.norm().ppf(rdpoints)
        L = cholesky(corr, lower=True)

        norm_u = np.matmul(norm_u, L)

        H = stats.norm().cdf(norm_u)
    else:
        H = np.zeros_like(rdpoints, dtype=float)
        rank = np.argsort(rdpoints, axis=0)

        for l in range(samples):
            low = float(l) / samples
            high = float(l + 1) / samples

            l_pos = rank == l
            H[l_pos] = random_state.uniform(low, high, size=N)
    return H


################################################################################

def _pdist(x):
    """
    Calculate the pair-wise point distances of a matrix

    Parameters
    ----------
    x : 2d-array
        An m-by-n array of scalars, where there are m points in n dimensions.

    Returns
    -------
    d : array
        A 1-by-b array of scalars, where b = m*(m - 1)/2. This array contains
        all the pair-wise point distances, arranged in the order (1, 0),
        (2, 0), ..., (m-1, 0), (2, 1), ..., (m-1, 1), ..., (m-1, m-2).

    Examples
    --------
    ::

        >>> x = np.array([[0.1629447, 0.8616334],
        ...               [0.5811584, 0.3826752],
        ...               [0.2270954, 0.4442068],
        ...               [0.7670017, 0.7264718],
        ...               [0.8253975, 0.1937736]])
        >>> _pdist(x)
        array([ 0.6358488,  0.4223272,  0.6189940,  0.9406808,  0.3593699,
                0.3908118,  0.3087661,  0.6092392,  0.6486001,  0.5358894])

    """

    x = np.atleast_2d(x)
    assert len(x.shape) == 2, 'Input array must be 2d-dimensional'

    m, n = x.shape
    if m < 2:
        return []

    d = []
    for i in range(m - 1):
        for j in range(i + 1, m):
            d.append((sum((x[j, :] - x[i, :]) ** 2)) ** 0.5)

    return np.array(d)


def ccdesign(n, center=(4, 4), alpha='orthogonal', face='circumscribed'):
    """
    Central composite design

    Parameters
    ----------
    n : int
        The number of factors in the design.

    Optional
    --------
    center : int array
        A 1-by-2 array of integers, the number of center points in each block
        of the design. (Default: (4, 4)).
    alpha : str
        A string describing the effect of alpha has on the variance. ``alpha``
        can take on the following values:

        1. 'orthogonal' or 'o' (Default)

        2. 'rotatable' or 'r'

    face : str
        The relation between the start points and the corner (factorial) points.
        There are three options for this input:

        1. 'circumscribed' or 'ccc': This is the original form of the central
           composite design. The star points are at some distance ``alpha``
           from the center, based on the properties desired for the design.
           The start points establish new extremes for the low and high
           settings for all factors. These designs have circular, spherical,
           or hyperspherical symmetry and require 5 levels for each factor.
           Augmenting an existing factorial or resolution V fractional
           factorial design with star points can produce this design.

        2. 'inscribed' or 'cci': For those situations in which the limits
           specified for factor settings are truly limits, the CCI design
           uses the factors settings as the star points and creates a factorial
           or fractional factorial design within those limits (in other words,
           a CCI design is a scaled down CCC design with each factor level of
           the CCC design divided by ``alpha`` to generate the CCI design).
           This design also requires 5 levels of each factor.

        3. 'faced' or 'ccf': In this design, the star points are at the center
           of each face of the factorial space, so ``alpha`` = 1. This
           variety requires 3 levels of each factor. Augmenting an existing
           factorial or resolution V design with appropriate star points can
           also produce this design.

    Notes
    -----
    - Fractional factorial designs are not (yet) available here.
    - 'ccc' and 'cci' can be rotatable design, but 'ccf' cannot.
    - If ``face`` is specified, while ``alpha`` is not, then the default value
      of ``alpha`` is 'orthogonal'.

    Returns
    -------
    mat : 2d-array
        The design matrix with coded levels -1 and 1

    Example
    -------
    ::

        >>> ccdesign(3)
        array([[-1.        , -1.        , -1.        ],
               [ 1.        , -1.        , -1.        ],
               [-1.        ,  1.        , -1.        ],
               [ 1.        ,  1.        , -1.        ],
               [-1.        , -1.        ,  1.        ],
               [ 1.        , -1.        ,  1.        ],
               [-1.        ,  1.        ,  1.        ],
               [ 1.        ,  1.        ,  1.        ],
               [ 0.        ,  0.        ,  0.        ],
               [ 0.        ,  0.        ,  0.        ],
               [ 0.        ,  0.        ,  0.        ],
               [ 0.        ,  0.        ,  0.        ],
               [-1.82574186,  0.        ,  0.        ],
               [ 1.82574186,  0.        ,  0.        ],
               [ 0.        , -1.82574186,  0.        ],
               [ 0.        ,  1.82574186,  0.        ],
               [ 0.        ,  0.        , -1.82574186],
               [ 0.        ,  0.        ,  1.82574186],
               [ 0.        ,  0.        ,  0.        ],
               [ 0.        ,  0.        ,  0.        ],
               [ 0.        ,  0.        ,  0.        ],
               [ 0.        ,  0.        ,  0.        ]])


    """
    # Check inputs
    assert isinstance(n, int) and n > 1, '"n" must be an integer greater than 1.'
    assert alpha.lower() in ('orthogonal', 'o', 'rotatable',
                             'r'), 'Invalid value for "alpha": {:}'.format(alpha)
    assert face.lower() in ('circumscribed', 'ccc', 'inscribed', 'cci',
                            'faced', 'ccf'), 'Invalid value for "face": {:}'.format(face)

    try:
        nc = len(center)
    except:
        raise TypeError('Invalid value for "center": {:}. Expected a 1-by-2 array.'.format(center))
    else:
        if nc != 2:
            raise ValueError('Invalid number of values for "center" (expected 2, but got {:})'.format(nc))

    # Orthogonal Design
    if alpha.lower() in ('orthogonal', 'o'):
        H2, a = star(n, alpha='orthogonal', center=center)

    # Rotatable Design
    if alpha.lower() in ('rotatable', 'r'):
        H2, a = star(n, alpha='rotatable')

    # Inscribed CCD
    if face.lower() in ('inscribed', 'cci'):
        H1 = ff2n(n)
        H1 = H1 / a  # Scale down the factorial points
        H2, a = star(n)

    # Faced CCD
    if face.lower() in ('faced', 'ccf'):
        H2, a = star(n)  # Value of alpha is always 1 in Faced CCD
        H1 = ff2n(n)

    # Circumscribed CCD
    if face.lower() in ('circumscribed', 'ccc'):
        H1 = ff2n(n)

    C1 = repeat_center(n, center[0])
    C2 = repeat_center(n, center[1])

    H1 = union(H1, C1)
    H2 = union(H2, C2)
    H = union(H1, H2)

    return H


def ff2n(n):
    """
    Create a 2-Level full-factorial design

    Parameters
    ----------
    n : int
        The number of factors in the design.

    Returns
    -------
    mat : 2d-array
        The design matrix with coded levels -1 and 1

    Example
    -------
    ::

        >>> ff2n(3)
        array([[-1., -1., -1.],
               [ 1., -1., -1.],
               [-1.,  1., -1.],
               [ 1.,  1., -1.],
               [-1., -1.,  1.],
               [ 1., -1.,  1.],
               [-1.,  1.,  1.],
               [ 1.,  1.,  1.]])

    """
    return 2 * fullfact([2] * n) - 1


def union(H1, H2):
    """
    Join two matrices by stacking them on top of each other.

    Parameters
    ----------
    H1 : 2d-array
        The matrix that goes on top of the new matrix
    H2 : 2d-array
        The matrix that goes on bottom of the new matrix

    Returns
    -------
    mat : 2d-array
        The new matrix that contains the rows of ``H1`` on top of the rows of
        ``H2``.

    Example
    -------
    ::

        >>> union(np.eye(2), -np.eye(2))
        array([[ 1.,  0.],
               [ 0.,  1.],
               [-1.,  0.],
               [ 0., -1.]])

    """
    H = np.r_[H1, H2]
    return H


def star(n, alpha='faced', center=(1, 1)):
    """
    Create the star points of various design matrices

    Parameters
    ----------
    n : int
        The number of variables in the design

    Optional
    --------
    alpha : str
        Available values are 'faced' (default), 'orthogonal', or 'rotatable'
    center : array
        A 1-by-2 array of integers indicating the number of center points
        assigned in each block of the response surface design. Default is
        (1, 1).

    Returns
    -------
    H : 2d-array
        The star-point portion of the design matrix (i.e. at +/- alpha)
    a : scalar
        The alpha value to scale the star points with.

    Example
    -------
    ::

        >>> star(3)
        array([[-1.,  0.,  0.],
               [ 1.,  0.,  0.],
               [ 0., -1.,  0.],
               [ 0.,  1.,  0.],
               [ 0.,  0., -1.],
               [ 0.,  0.,  1.]])

    """
    # Star points at the center of each face of the factorial
    if alpha == 'faced':
        a = 1
    elif alpha == 'orthogonal':
        nc = 2 ** n  # factorial points
        nco = center[0]  # center points to factorial
        na = 2 * n  # axial points
        nao = center[1]  # center points to axial design
        # value of alpha in orthogonal design
        a = (n * (1 + nao / float(na)) / (1 + nco / float(nc))) ** 0.5
    elif alpha == 'rotatable':
        nc = 2 ** n  # number of factorial points
        a = nc ** (0.25)  # value of alpha in rotatable design
    else:
        raise ValueError('Invalid value for "alpha": {:}'.format(alpha))

    # Create the actual matrix now.
    H = np.zeros((2 * n, n))
    for i in range(n):
        H[2 * i:2 * i + 2, i] = [-1, 1]

    H *= a

    return H, a


def build_gsd(levels, reduction, n=1):
    """
    Create a Generalized Subset Design (GSD).
    Parameters
    ----------
    levels : array-like
        Number of factor levels per factor in design.
    reduction : int
        Reduction factor (bigger than 1). Larger `reduction` means fewer
        experiments in the design and more possible complementary designs.
    n : int
        Number of complementary GSD-designs (default 1). The complementary
        designs are balanced analogous to fold-over in two-level fractional
        factorial designs.
    Returns
    -------
    H : 2d-array | list of 2d-arrays
        `n` m-by-k matrices where k is the number of factors (equal
        to the length of `factor_levels`. The number of rows, m, will
        be approximately equal to the grand product of the factor levels
        divided by `reduction`.
    Raises
    ------
    ValueError
        If input is valid or if design construction fails. Design can fail
        if `reduction` is too large compared to values of `levels`.
    Notes
    -----
    The Generalized Subset Design (GSD) [1]_ or generalized factorial design is
    a generalization of traditional fractional factorial designs to problems
    where factors can have more than two levels.
    In many application problems factors can have categorical or quantitative
    factors on more than two levels. Previous reduced designs have not been
    able to deal with such types of problems. Full multi-level factorial
    designs can handle such problems but are however not economical regarding
    the number of experiments.
    Note for commercial users, the application of GSD to testing of product
    characteristics in a processing facility is patented [2]_
    Examples
    --------
    An example with three factors using three, four and
    six levels respectively reduced with a factor 4 ::
        >>> gsd([3, 4, 6], 4)
        array([[0, 0, 0],
               [0, 0, 4],
               [0, 1, 1],
               [0, 1, 5],
               [0, 2, 2],
               [0, 3, 3],
               [1, 0, 1],
               [1, 0, 5],
               [1, 1, 2],
               [1, 2, 3],
               [1, 3, 0],
               [1, 3, 4],
               [2, 0, 2],
               [2, 1, 3],
               [2, 2, 0],
               [2, 2, 4],
               [2, 3, 1],
               [2, 3, 5]])
    Two complementary designs with two factors using three and
    four levels reduced with a factor 2 ::
        >>> gsd([3, 4], 2, n=2)[0]
        array([[0, 0],
               [0, 2],
               [2, 0],
               [2, 2],
               [1, 1],
               [1, 3]])
        >>> gsd([3, 4], 2, n=2)[1]
        array([[0, 1],
               [0, 3],
               [2, 1],
               [2, 3],
               [1, 0],
               [1, 2]])
    If design fails ValueError is raised ::
        >>> gsd([2, 3], 5)
        Traceback (most recent call last):
         ...
        ValueError: reduction too large compared to factor levels
    References
    ----------
    .. [1] Surowiec, Izabella, Ludvig Vikstrom, Gustaf Hector, Erik Johansson,
       Conny Vikstrom, and Johan Trygg. "Generalized Subset Designs in
       Analytical Chemistry." Analytical Chemistry 89, no. 12 (June 20, 2017):
       6491-97. https://doi.org/10.1021/acs.analchem.7b00506.
    .. [2] Vikstrom, Ludvig, Conny Vikstrom, Erik Johansson, and Gustaf Hector.
       Computer-implemented systems and methods for generating
       generalized fractional designs. US9746850 B2, filed May 9,
       2014, and issued August 29, 2017. http://www.google.se/patents/US9746850.
    """
    try:
        assert all(isinstance(v, int) for v in levels), \
            'levels has to be sequence of integers'
        assert isinstance(reduction, int) and reduction > 1, \
            'reduction has to be integer larger than 1'
        assert isinstance(n, int) and n > 0, \
            'n has to be positive integer'
    except AssertionError as e:
        raise ValueError(e)

    partitions = _make_partitions(levels, reduction)
    latin_square = _make_latin_square(reduction)
    ortogonal_arrays = _make_orthogonal_arrays(latin_square, len(levels))

    try:
        designs = [_map_partitions_to_design(partitions, oa) - 1 for oa in
                   ortogonal_arrays]
    except ValueError:
        raise ValueError('reduction too large compared to factor levels')

    if n == 1:
        return designs[0]
    else:
        return designs[:n]


def _make_orthogonal_arrays(latin_square, n_cols):
    """
    Augment latin-square to the specified number of columns to produce
    an orthogonal array.
    """
    p = len(latin_square)

    first_row = latin_square[0]
    A_matrices = [np.array([[v]]) for v in first_row]

    while A_matrices[0].shape[1] < n_cols:
        new_A_matrices = list()

        for i, A_matrix in enumerate(A_matrices):
            sub_a = list()
            for constant, other_A in zip(first_row,
                                         np.array(A_matrices)[latin_square[i]]):
                constant_vec = np.repeat(constant, len(other_A))[:, np.newaxis]
                combined = np.hstack([constant_vec, other_A])
                sub_a.append(combined)

            new_A_matrices.append(np.vstack(sub_a))

        A_matrices = new_A_matrices

        if A_matrices[0].shape[1] == n_cols:
            break

    return A_matrices


def _map_partitions_to_design(partitions, ortogonal_array):
    """
    Map partitioned factor to final design using orthogonal-array produced
    by augmenting latin square.
    """
    assert len(
        partitions) == ortogonal_array.max() + 1 and ortogonal_array.min() == 0, \
        'Orthogonal array indexing does not match partition structure'

    mappings = list()
    for row in ortogonal_array:
        if any(not partitions[p][factor] for factor, p in enumerate(row)):
            continue

        partition_sets = [partitions[p][factor] for factor, p in enumerate(row)]
        mapping = list(itertools.product(*partition_sets))
        mappings.append(mapping)

    return np.vstack(mappings)


def _make_partitions(factor_levels, num_partitions):
    """
    Balanced partitioning of factors.
    """
    partitions = list()
    for partition_i in range(1, num_partitions + 1):
        partition = list()

        for num_levels in factor_levels:
            part = list()
            for level_i in range(1, num_levels):
                index = partition_i + (level_i - 1) * num_partitions
                if index <= num_levels:
                    part.append(index)

            partition.append(part)

        partitions.append(partition)

    return partitions


def _make_latin_square(n):
    numbers = np.arange(n)
    latin_square = np.vstack([np.roll(numbers, -i) for i in range(n)])
    return latin_square


def build_halton(factor_level_ranges, num_samples=None):
    """
    Builds a quasirandom dataframe from a dictionary of factor/level ranges using prime numbers as seed.
    Only min and max values of the range are required.
    Example of the dictionary which is needed as the input:
    {'Pressure':[50,70],'Temperature':[290, 350],'Flow rate':[0.9,1.0]}
    num_samples: Number of samples to be generated
    Quasirandom sequence using the default initialization with first n prime numbers equal to the number of factors/variables.
    """
    for key in factor_level_ranges:
        if len(factor_level_ranges[key]) != 2:
            factor_level_ranges[key][1] = factor_level_ranges[key][-1]
            factor_level_ranges[key] = factor_level_ranges[key][:2]
            print(
                f"{key} had more than two levels. Assigning the end point to the high level."
            )

    factor_count = len(factor_level_ranges)
    factor_lists = []

    if num_samples == None:
        num_samples = factor_count

    for key in factor_level_ranges:
        factor_lists.append(factor_level_ranges[key])

    x = halton(
        num_points=num_samples, dimension=factor_count
    )  # create Halton matrix design
    factor_lists = np.array(factor_lists)

    df = construct_df_from_random_matrix(x, factor_lists)
    # df.columns = factor_level_ranges.keys()
    return df


def halton(num_points, dimension):
    """Halton sequence.
    :param int dim: dimension
    :param int n_sample: number of samples.
    :return: sequence of Halton.
    :rtype: array_like (n_samples, n_features)
    """
    big_number = 10
    while 'Not enought primes':
        base = _primes_from_2_to(big_number)[:dimension]
        if len(base) == dimension:
            break
        big_number += 1000

    # Generate a sample using a Van der Corput sequence per dimension.
    sample = [_van_der_corput(num_points + 1, dimension) for dimension in base]
    sample = np.stack(sample, axis=-1)[1:]

    return sample


def _primes_from_2_to(n):
    """Prime number from 2 to n.
    From `StackOverflow <https://stackoverflow.com/questions/2068372>`_.
    :param int n: sup bound with ``n >= 6``.
    :return: primes in 2 <= p < n.
    :rtype: list
    """
    sieve = np.ones(n // 3 + (n % 6 == 2), dtype=np.bool)
    for i in range(1, int(n ** 0.5) // 3 + 1):
        if sieve[i]:
            k = 3 * i + 1 | 1
            sieve[k * k // 3::2 * k] = False
            sieve[k * (k - 2 * (i & 1) + 4) // 3::2 * k] = False
    return np.r_[2, 3, ((3 * np.nonzero(sieve)[0][1:] + 1) | 1)]


def _van_der_corput(n_sample, base=2):
    """Van der Corput sequence.
    :param int n_sample: number of element of the sequence.
    :param int base: base of the sequence.
    :return: sequence of Van der Corput.
    :rtype: list (n_samples,)
    """
    sequence = []
    for i in range(n_sample):
        n_th_number, denom = 0., 1.
        while i > 0:
            i, remainder = divmod(i, base)
            denom *= base
            n_th_number += remainder / denom
        sequence.append(n_th_number)

    return sequence



import unittest

from ..problem import Problem
from ..algorithm_sensitivity import Sensitivity, SALibAlgorithm

import math


class SensitivitySALibProblem(Problem):
    def set(self):
        self.name = "SensitivityTestProblem"
        self.parameters = [{'name': 'A', 'initial_value': 0.0, 'bounds': [-3.14159, 3.14159]},
                           {'name': 'B', 'initial_value': 0.0, 'bounds': [-3.14159, 3.14159]},
                           {'name': 'C', 'initial_value': 0.0, 'bounds': [-3.14159, 3.14159]}]

        self.costs = [{'name': 'F_1'}]
        # self.executor = CondorComsolJobExecutor(self, model_file="./data/coil_linear_current.mph",
        #                                        files_from_condor=["temp.txt"])

    def evaluate(self, individual):
        return [math.sin(individual.vector[0]) + 7.0 * math.pow(math.sin(individual.vector[1]), 2) + \
                0.1 * math.pow(individual.vector[2], 4) * math.sin(individual.vector[0])]


class TestSALibSensitivity(unittest.TestCase):
    """ Tests simple one objective optimization problem."""

    def test_sobol_problem(self):
        problem = SensitivitySALibProblem()
        problem.options['save_data_files'] = True

        algorithm = SALibAlgorithm(problem)
        algorithm.options["samples"] = 1000
        algorithm.options["method"] = "sobol"
        algorithm.options["print_to_console"] = False
        algorithm.run()

        res = algorithm.analyze()
        self.assertAlmostEqual(res["S1"][0], 0.30797525)
        self.assertAlmostEqual(res["S1"][1], 0.44776754)
        self.assertAlmostEqual(res["S1"][2], -0.00425456)


class MyProblem(Problem):
    """ Describe simple one objective optimization problem. """
    def set(self):
        self.name = "LocalPythonProblem"
        self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [0, 5], 'precision': 1e-1},
                           {'name': 'x_2', 'initial_value': 2.5, 'bounds': [2.2, 2.4], 'precision': 1e-1},
                           {'name': 'x_3', 'initial_value': 2.5, 'bounds': [0, 5], 'precision': 1e-1}]
        self.costs = [{'name': 'F'}]

    def evaluate(self, individual):
        x = individual.vector
        result = x[0] * x[0] + 0.1 * x[1] * x[1] + x[2] * x[2]
        return [result]


class TestSensitivity(unittest.TestCase):
    """ Tests simple one objective optimization problem."""

    def test_local_problem(self):
        problem = MyProblem()
        algorithm = Sensitivity(problem, ['x_2', 'x_3'])
        algorithm.options['max_population_size'] = 10
        algorithm.run()


if __name__ == '__main__':
    unittest.main()



import math
import unittest

from ..problem import Problem
from ..individual import Individual
from ..benchmark_functions import Booth
from ..surrogate_smt import SurrogateModelSMT

from smt.surrogate_models import RBF, KRG, KPLS, IDW, RMTB, LS, QP, GENN, GEKPLS, KPLSK


class TestSurrogateSMT(unittest.TestCase):
    def _check_smt(self, regressor, percent=1.0):
        problem = Booth()
        problem.surrogate = SurrogateModelSMT(problem)
        # set custom regressor
        problem.surrogate.regressor = regressor
        # set threshold
        problem.surrogate.train_step = 12

        # train
        problem.surrogate.evaluate(Individual([1.0, 3.1]))
        problem.surrogate.evaluate(Individual([1.3, 3.2]))
        problem.surrogate.evaluate(Individual([1.8, 2.7]))
        problem.surrogate.evaluate(Individual([0.9, 3.3]))
        problem.surrogate.evaluate(Individual([0.8, 3.1]))
        problem.surrogate.evaluate(Individual([1.4, 2.8]))
        problem.surrogate.evaluate(Individual([0.1, 3.0]))
        problem.surrogate.evaluate(Individual([0.2, 3.4]))
        problem.surrogate.evaluate(Individual([0.95, 3.03]))
        problem.surrogate.evaluate(Individual([0.98, 2.96]))
        problem.surrogate.evaluate(Individual([1.02, 2.98]))
        problem.surrogate.evaluate(Individual([1.01, 2.99]))

        # check values
        x_ref = Individual([0.2, 3.1])
        # eval reference
        value_problem = problem.evaluate(x_ref)[0]
        # eval surrogate
        value_surrogate = problem.surrogate.predict(x_ref.vector)

        percent_check = 100.0 * math.fabs(value_problem - value_surrogate) / math.fabs(value_problem)
        problem.logger.info(
            "{}: surrogate.value: eval = {}, pred = {}, diff = {} ({} %)".format(problem.name, value_problem,
                                                                                 value_surrogate,
                                                                                 math.fabs(
                                                                                     value_problem - value_surrogate),
                                                                                 percent_check))

        self.assertLess(percent_check, percent)

    def test_smt_rbf(self):
        self._check_smt(RBF(d0=5, print_prediction=False))

    #def test_smt_idw(self):
    #    self._check_smt(IDW(p=4))

    def test_smt_kpls(self):
        self._check_smt(KPLS(theta0=[1e-2]))

    def test_smt_kplsk(self):
        self._check_smt(KPLSK(theta0=[1e-2]))

    #def test_smt_ls(self):
    #    self._check_smt(LS())

    def test_smt_qp(self):
        self._check_smt(QP())

    def test_smt_krg(self):
        self._check_smt(KRG(theta0=[1e-2]))



from scipy import spatial
import numpy as np

"""
Contains different quality indicators to benchmark the multi-objective optimization algorithms.

Indicators:

    - Generational Distance
    - Additive unary epsilon indicator
"""


def gd(reference: list, computed: list, norm='euclidean'):
    """
    :param reference: list of the reference points [(coord_1, coord_2, coord_3,..., coord_n)]
    :return: the sum of the euclidean disctance from the given reference points.

    Possible norms:  ‘braycurtis’, ‘canberra’, ‘chebyshev’, ‘cityblock’, ‘correlation’, ‘cosine’, ‘dice’,
                     ‘euclidean’, ‘hamming’, ‘jaccard’, ‘jensenshannon’, ‘kulsinski’, ‘mahalanobis’, ‘matching’,
                     ‘minkowski’, ‘rogerstanimoto’, ‘russellrao’, ‘seuclidean’, ‘sokalmichener’, ‘sokalsneath’,
                     ‘sqeuclidean’, ‘wminkowski’, ‘yule’.

    Type: Convergence Indicator

    The generational distance performence indicator [1,2,3] measueres the distancee from the
    given reference list. T

    .. math::

        GD(S,P) = \frac{1}{n}(\Sigma_{s \in S} min ||F(s) - F(r)||^2)^{\frac{1}{p}}

    where |S| is the number of the points in a Pareto-set approximation and P a discrete representation
    of the Pareto-front. Generally, p=2. In this case it is equivalent with M1 measure [4].
    When P = 1 it is equivalent with the Gamma-Metric.

    References
    ----------

    .: [1] David A. Van Veldhuizen and David A. Van Veldhuizen.
           Multi-objective evolutionary algorithms: classifications, analyses, and new innovations.
           Technical Report, Evolutionary Computation, 1999.
    .: [2] Audet, Charles, J. Bigeon, D. Cartier, Sébastien Le Digabel, and Ludovic Salomon.
           "Performance indicators in multiobjective optimization." Optimization Online (2018).
    .: [3] https://pymoo.org/misc/performance_indicator.html
    .: [4] E. Zitzler, K. Deb, L. Thiele, Comparison of multiobjective evolutionary algorithms: Empirical
           results, Evolutionary computation 8 (2) (2000) 173–195.

    """

    distances = spatial.distance.cdist(reference, computed, metric=norm)
    minimums = np.nanmin(distances, axis=0)

    return np.sum(minimums) / len(computed)


def epsilon_add(reference: list, computed: list):
    """
     :param reference: list of reference pareto-front values, list of tuples
     :param computed: list of the computed pareto-front values, list of tuples
     :return: additive unary epsilon indicator

     .. math::
        I_{\epsilon}(A,B) = max_{x^2 \in B} min_{x^1 \in A} max_{1 \leq i \leq m} f_i(x^1) - f_i(x^2)

     [1] E. Zitzler, E. Thiele, L. Laummanns, M., Fonseca, C., and Grunert da Fonseca. V (2003):
        Performance Assessment of Multiobjective Optimizers: An Analysis and
        Review. The code is the a Java version of the original metric implementation by Eckart Zitzler.

     [2] Audet, Charles, J. Bigeon, D. Cartier, Sébastien Le Digabel, and Ludovic Salomon.
         "Performance indicators in multiobjective optimization." Optimization Online (2018).
    """

    eps = 0.0
    for ref_val in reference:
        eps_j = np.infty
        for comp_val in computed:
            eps_k = max(np.subtract(comp_val, ref_val))
            eps_j = min(eps_k, eps_j)
        eps = max(eps, eps_j)

    return eps




"""
 Module is dedicated to describe optimization problem.
"""

from datastore import SqliteDataStore, DummyDataStore
from utils import ConfigDictionary
from artap.surrogate import SurrogateModelEval
from abc import abstractmethod

import logging
import datetime
import tempfile
import os
import shutil
import atexit
import platform

CRITICAL = logging.CRITICAL
FATAL = logging.FATAL
ERROR = logging.ERROR
WARNING = logging.WARNING
WARN = logging.WARN
INFO = logging.INFO
DEBUG = logging.DEBUG
NOTSET = logging.NOTSET

_log_level = [CRITICAL, ERROR, WARNING, INFO, DEBUG]


class Problem:
    """ The Class Problem Is a main class which collects information about optimization task """

    # __is_frozen = False

    def __init__(self, **kwargs):

        self.options = ConfigDictionary()
        # options
        self.options.declare(name='calculate_gradients', default=False,
                             desc='calculate gradient for individuals')
        self.options.declare(name='log_level', default=logging.INFO, values=_log_level,
                             desc='Log level')
        self.options.declare(name='log_file_handler', default=False,
                             desc='Enable file handler')
        self.options.declare(name='log_db_handler', default=True,
                             desc='Enable DB handler')
        self.options.declare(name='log_console_handler', default=True,
                             desc='Enable console handler')
        self.options.declare(name='time_out', default=600,
                             desc='Maximal time for calculation')
        self.options.declare(name='save_data_files', default=False,
                             desc='Saving data from computation')

        # tmp name
        d = datetime.datetime.now()
        ts = d.strftime("{}-%f".format(self.__class__.__name__))

        # create logger
        self.logger = logging.getLogger(ts)
        self.logger.setLevel(self.options['log_level'])

        # create formatter
        self.formatter = logging.Formatter('%(asctime)s (%(levelname)s): '
                                           '%(name)s - %(funcName)s (%(lineno)d) - %(message)s')

        if self.options['log_console_handler']:
            # create console handler and set level to debug
            stream_err_handler = logging.StreamHandler()
            stream_err_handler.setLevel(logging.DEBUG)
            # add formatter to StreamHandler
            stream_err_handler.setFormatter(self.formatter)
            # add StreamHandler to logger
            self.logger.addHandler(stream_err_handler)

            # stream_std_handler = logging.StreamHandler()
            # stream_std_handler.setLevel(logging.INFO)
            # # add formatter to StreamHandler
            # stream_std_handler.setFormatter(self.formatter)
            # # add StreamHandler to logger
            # self.logger.addHandler(stream_std_handler)

        # for h in list(self.logger.handlers):
        #     print(h)

        self.logger.debug("START")

        self.name: str = str()
        self.description: str = str()
        self.parameters: list = list()
        self.costs: list = list()
        self.signs = []

        self.info = {"processor": platform.processor(),
                     "system": platform.system(),
                     "python": platform.python_version(),
                     "hostname": platform.node()}


        # populations
        self.individuals = []
        self.data_store = None
        self.executor = None
        self.failed = []  # storage for failed individuals
        self.data_store = DummyDataStore()
        self.output_files = None

        self.working_dir = tempfile.gettempdir() + os.sep + "artap-{}".format(ts) + os.sep
        os.mkdir(self.working_dir)

        # surrogate model (default - only simple eval)
        self.surrogate = SurrogateModelEval(self)

        # self._freeze()
        self.set(**kwargs)
        for cost in self.costs:
            if 'criteria' in cost:
                if cost['criteria'] == 'minimize':
                    self.signs.append(1)
                else:
                    self.signs.append(-1)
            else:
                self.signs.append(1)

        # clean up
        atexit.register(self.cleanup)

    def populations(self):
        individuals = {}
        for individual in self.individuals:
            if individual.population_id not in individuals:
                individuals[individual.population_id] = []
            individuals[individual.population_id].append(individual)

        return individuals

    def population(self, population_id):
        individuals = []
        for individual in self.individuals:
            if individual.population_id == population_id:
                individuals.append(individual)

        return individuals

    def last_population(self):
        # TODO: temporary solution - faster implementation with TinyDB

        # find max index
        max_index = -1
        for individual in self.individuals:
            if individual.population_id > max_index:
                max_index = individual.population_id

        # add to population
        return self.population(max_index)

    def to_dict(self):
        parameters = list(self.parameters)
        output = {'name': self.name, 'description': self.description, 'parameters': parameters}
        return output

    @staticmethod
    def from_dict():
        pass

    def cleanup(self):
        if self.data_store:
            self.data_store.destroy()

        if self.working_dir.startswith(tempfile.gettempdir() + os.sep + "artap-"):
            shutil.rmtree(self.working_dir)

    @abstractmethod
    def set(self, **kwargs):
        pass

    def parameters_len(self):
        return len(self.parameters)

    # search for and defines the initial values for the optimization problem
    def get_initial_values(self):
        values = []
        for parameter in self.parameters:
            if 'initial_value' in parameter:
                values.append(parameter['initial_value'])
            else:
                values.append(0)
        return values

    def get_parameter_types(self):
        p_types = []
        for parameter in self.parameters:

            if 'parameter_type' in parameter:
                if parameter['parameter_type'].lower == 'real':
                    p_types.append('real')

                if parameter['parameter_type'].lower == 'integer':
                    p_types.append('integer')

                if parameter['parameter_type'].lower == 'boolean':
                    p_types.append('boolean')

            else:
                p_types.append('real')

        return p_types

    @abstractmethod
    def evaluate(self, individual):
        """ :param individual: Individual """
        pass

    def evaluate_constraints(self, individual):
        """ :param individual: Individual """
        pass

    def __setattr__(self, key, value):
        # if self.__is_frozen and not hasattr(self, key):
        #     raise TypeError(" %r is a frozen class" % self)
        object.__setattr__(self, key, value)

        # working dir must be set
        if hasattr(self.options, 'log_file_handler'):
            if self.working_dir:
                # create file handler and set level to debug
                file_handler = logging.FileHandler(self.working_dir + "/data.log")
                file_handler.setLevel(logging.DEBUG)
                # add formatter to FileHandler
                file_handler.setFormatter(self.formatter)
                # add FileHandler to logger
                self.logger.addHandler(file_handler)

    # def _freeze(self):
    #     self.__is_frozen = True


class ProblemViewDataStore(Problem):
    def __init__(self, database_name):
        super().__init__()
        self.data_store = SqliteDataStore(self, database_name=database_name, mode="read")

    def set(self, **kwargs):
        pass

    def evaluate(self, individual):
        pass



import os
import unittest
import tempfile
import time
import json
import sqlite3
import pathlib

from ..problem import Problem, ProblemViewDataStore
from ..individual import Individual
from ..datastore import SqliteDataStore
from ..algorithm_sweep import SweepAlgorithm
from ..algorithm_genetic import NSGAII

from ..operators import RandomGenerator

from sys import platform
if platform == "win32":
    __platform__ = 'WINDOWS'
else:
    __platform__ = 'Linux'


class MyProblem(Problem):
    """ Describe simple one objective optimization problem. """
    def set(self):
        self.name = "NLopt_BOBYQA"
        self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [-10, 10]},
                           {'name': 'x_2', 'initial_value': 1.5, 'bounds': [-10, 10]}]
        self.costs = [{'name': 'F', 'criteria': 'minimize'}]

    def evaluate(self, individual):
        x_1 = individual.vector[0]
        x_2 = individual.vector[1]

        # set custom properties
        individual.custom["functions"] = [x_1**2, x_2**2]

        return [x_1**2 + x_2**2]


class TestDataStoreSqlite(unittest.TestCase):
    def test_read_write_database(self):
        problem = MyProblem()
        # set data store
        # database_name = tempfile.NamedTemporaryFile(mode="w", delete=False, dir=None, suffix=".sqlite").name
        database_name = "database.sqlite"
        problem.data_store = SqliteDataStore(problem, database_name=database_name)

        algorithm = NSGAII(problem)
        algorithm.options['max_population_number'] = 6
        algorithm.options['max_population_size'] = 4
        algorithm.options['max_processes'] = 4
        algorithm.run()

        # cache individuals
        individuals = {}
        for individual in problem.individuals:
            individuals[individual.id] = individual

        # remove datastore
        problem.data_store.destroy()

        # check json
        conn = sqlite3.connect(database_name)
        c = conn.cursor()
        c.execute("SELECT * FROM individuals WHERE ID = ?", [list(individuals.keys())[6]])
        row = c.fetchall()
        individual = Individual.from_dict(json.loads(row[0][1]))
        # print(individual)

        # result
        individual_orig = individuals[individual.id]
        self.assertAlmostEqual(individual.costs[0], individual_orig.costs[0], 3)
        self.assertAlmostEqual(individual.vector[0], individual_orig.vector[0], 3)
        self.assertAlmostEqual(individual.vector[1], individual_orig.vector[1], 3)
        for key, value in individual_orig.custom.items():
            self.assertEqual(value, individual.custom[key])
        for key, value in individual_orig.features.items():
            self.assertEqual(value, individual.features[key])

        # remove file
        # os.remove(database_name)

    def test_read_datastore(self):
        # Path to this script file location
        file_path = str(pathlib.Path(__file__).parent.absolute())
        database_name = os.path.join(file_path, "data/data.sqlite")
        problem = ProblemViewDataStore(database_name=database_name)

        self.assertEqual(problem.name, 'NLopt_BOBYQA')

        individuals = problem.last_population()
        self.assertAlmostEqual(problem.parameters[0]['bounds'][0], -10)
        self.assertAlmostEqual(individuals[0].vector[1], 6.58962945, 4)
        self.assertAlmostEqual(individuals[1].vector[0], 1.8944488, 4)
        self.assertAlmostEqual(individuals[0].costs[0], 49.0245242, 4)


class TestDataStoreBenchmark(unittest.TestCase):
    def setUp(self):
        self.n = 300
        self.max_processes = 10

    def test_benchmark_sqlite_data_store(self):
        t_s = time.time()
        problem = MyProblem()

        # set data store
        database_name = tempfile.NamedTemporaryFile(mode="w", delete=False, dir=None, suffix=".sqlite").name
        problem.data_store = SqliteDataStore(problem, database_name=database_name, thread_safe=(self.max_processes > 1))

        gen = RandomGenerator(problem.parameters)
        gen.init(self.n)

        algorithm = SweepAlgorithm(problem, generator=gen)
        algorithm.options['max_processes'] = self.max_processes
        algorithm.run()

        # cache individuals
        individuals = {}
        for individual in problem.individuals:
            individuals[individual.id] = individual

        # sync
        problem.data_store.destroy()

        t = time.time() - t_s
        problem.logger.info("write elapsed time: {} s, size: {} MB, n: {}".format(t, self.get_size(database_name) / 1024 / 1024, self.n))

        # check db
        t_s = time.time()
        # check

        conn = sqlite3.connect(database_name)
        c = conn.cursor()
        c.execute("SELECT * FROM individuals;")
        rows = c.fetchall()
        for row in rows:
            individual = Individual.from_dict(json.loads(row[1]))
            self.assertAlmostEqual(individual.costs[0], individuals[individual.id].costs[0], 3)

        # remove file
        # print(database_name)
        # os.remove(database_name)

        t = time.time() - t_s
        problem.logger.info("read elapsed time: {} s".format(t))

    def get_size(self, start_path):
        if os.path.isfile(start_path):
            return os.path.getsize(start_path)
        else:
            total_size = 0
            for dirpath, dirnames, filenames in os.walk(start_path):
                for f in filenames:
                    fp = os.path.join(dirpath, f)
                    # skip if it is symbolic link
                    if not os.path.islink(fp):
                        total_size += os.path.getsize(fp)

            return total_size


if __name__ == '__main__':
    unittest.main()



import time
import nlopt

from .problem import Problem
from .algorithm import Algorithm

GN_DIRECT = nlopt.GN_DIRECT
GN_DIRECT_L = nlopt.GN_DIRECT_L
GN_DIRECT_L_RAND = nlopt.GN_DIRECT_L_RAND
GN_DIRECT_NOSCAL = nlopt.GN_DIRECT_NOSCAL
GN_DIRECT_L_NOSCAL = nlopt.GN_DIRECT_L_NOSCAL
GN_DIRECT_L_RAND_NOSCAL = nlopt.GN_DIRECT_L_RAND_NOSCAL
GN_ORIG_DIRECT = nlopt.GN_ORIG_DIRECT
GN_ORIG_DIRECT_L = nlopt.GN_ORIG_DIRECT_L
GD_STOGO = nlopt.GD_STOGO
GD_STOGO_RAND = nlopt.GD_STOGO_RAND
LD_LBFGS_NOCEDAL = nlopt.LD_LBFGS_NOCEDAL
LD_LBFGS = nlopt.LD_LBFGS
LN_PRAXIS = nlopt.LN_PRAXIS
LD_VAR1 = nlopt.LD_VAR1
LD_VAR2 = nlopt.LD_VAR2
LD_TNEWTON = nlopt.LD_TNEWTON
LD_TNEWTON_RESTART = nlopt.LD_TNEWTON_RESTART
LD_TNEWTON_PRECOND = nlopt.LD_TNEWTON_PRECOND
LD_TNEWTON_PRECOND_RESTART = nlopt.LD_TNEWTON_PRECOND_RESTART
GN_CRS2_LM = nlopt.GN_CRS2_LM
GN_MLSL = nlopt.GN_MLSL
GD_MLSL = nlopt.GD_MLSL
GN_MLSL_LDS = nlopt.GN_MLSL_LDS
GD_MLSL_LDS = nlopt.GD_MLSL_LDS
LD_MMA = nlopt.LD_MMA
LN_COBYLA = nlopt.LN_COBYLA
LN_NEWUOA = nlopt.LN_NEWUOA
LN_NEWUOA_BOUND = nlopt.LN_NEWUOA_BOUND
LN_NELDERMEAD = nlopt.LN_NELDERMEAD
LN_SBPLX = nlopt.LN_SBPLX
LN_AUGLAG = nlopt.LN_AUGLAG
LD_AUGLAG = nlopt.LD_AUGLAG
LN_AUGLAG_EQ = nlopt.LN_AUGLAG_EQ
LD_AUGLAG_EQ = nlopt.LD_AUGLAG_EQ
LN_BOBYQA = nlopt.LN_BOBYQA
GN_ISRES = nlopt.GN_ISRES
AUGLAG = nlopt.AUGLAG
AUGLAG_EQ = nlopt.AUGLAG_EQ
G_MLSL = nlopt.G_MLSL
G_MLSL_LDS = nlopt.G_MLSL_LDS
LD_SLSQP = nlopt.LD_SLSQP
LD_CCSAQ = nlopt.LD_CCSAQ
GN_ESCH = nlopt.GN_ESCH
GN_AGS = nlopt.GN_AGS

_algorithm = [nlopt.GN_DIRECT_L, GN_DIRECT_L_RAND, GN_MLSL, GN_CRS2_LM, GN_ISRES, GN_ESCH, LN_BOBYQA, LN_COBYLA, LN_NELDERMEAD, LN_SBPLX,  LN_PRAXIS, LN_AUGLAG_EQ]


class NLopt(Algorithm):
    """ NLopt algorithms """

    def __init__(self, problem: Problem, name="NLopt"):
        super().__init__(problem, name)

        self.options.declare(name='algorithm', default=LN_BOBYQA, values=_algorithm,
                             desc='Algorithm')
        self.options.declare(name='n_iterations', default=50, lower=1,
                             desc='Maximum evaluations')
        self.options.declare(name='xtol_rel', default=1e-6, lower=0.0,
                             desc='xtol_rel')
        self.options.declare(name='xtol_abs', default=1e-12, lower=0.0,
                             desc='xtol_abs')
        self.options.declare(name='ftol_rel', default=1e-6, lower=0.0,
                             desc='ftol_rel')
        self.options.declare(name='ftol_abs', default=1e-12, lower=0.0,
                             desc='ftol_abs')

    def _function(self, x, grad):
        return self.evaluator.evaluate_scalar(x)

    def _constraint(x, grad, a, b):
        # if grad.size > 0:
        #     grad[0] = 3 * a * (a * x[0] + b) ** 2
        #     grad[1] = -1.0
        # return (a * x[0] + b) ** 3 - x[1]
        return 0

    def run(self):
        # Figure out bounds vectors.
        lb = []
        ub = []
        for parameter in self.problem.parameters:
            bounds = parameter['bounds']

            lb.append(bounds[0])
            ub.append(bounds[1])

        op = nlopt.opt(self.options['algorithm'], len(self.problem.parameters))
        op.set_lower_bounds(lb)
        op.set_upper_bounds(ub)
        op.set_min_objective(self._function)
        op.set_xtol_rel(self.options['xtol_rel'])
        op.set_xtol_abs(self.options['xtol_abs'])
        op.set_ftol_rel(self.options['ftol_rel'])
        op.set_ftol_abs(self.options['ftol_abs'])
        op.set_maxeval(self.options['n_iterations'])

        # constraint
        # op.add_inequality_constraint(lambda x, grad: myconstraint(x, grad, 2, 0), 1e-8)
        # op.add_inequality_constraint(lambda x, grad: myconstraint(x, grad, -1, 1), 1e-8)

        try:
            t_s = time.time()

            self.problem.logger.info("NLopt: {}".format(op.get_algorithm_name()))
            x = op.optimize(self.problem.get_initial_values())
            # print('initial values:',x)

            t = time.time() - t_s
            self.problem.logger.info("NLopt: elapsed time: {} s".format(t))

            # sync changed individual informations
            self.problem.data_store.sync_all()

            """
            if self.options['verbose_level'] >= 1:
                print('method: ', op.get_algorithm_name())
                print('optimum at ', x)
                print('minimum value = ', op.last_optimum_value())
                print('nevals = ', op.get_numevals())
            """
        except RuntimeError:
            print('Optimization FAILED.')
            print(op.get_errmsg())
        except ValueError:
            print('Optimization FAILED.')
            print(op.get_errmsg())

        msg_nlopt = {-1: 'failure - generic failure code',
                     -2: 'failure - invalid arguments',
                     -3: 'failure - out of memory',
                     -4: 'failure - round off limited',
                     -5: 'failure - forced stop',
                      1: 'success - generic success code',
                      2: 'success - stop value reached',
                      3: 'success - ftol reached',
                      4: 'success - xtol reached',
                      5: 'success - maxeval reached',
                      6: 'success - maxtime reached'
                     }

        if self.options['verbose_level'] >= 1:
            print('optimum = ', op.last_optimum_value())
            print('result code and meaning = ', op.last_optimize_result(), msg_nlopt[op.last_optimize_result()])



import unittest

from ..problem import Problem
from ..algorithm_gradient import GradientAlgorithm


class ParabolicProblem(Problem):
    """ Describe simple one objective optimization problem. """

    def set(self):
        self.name = "LocalPythonProblem"
        self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [-5, 5], 'precision': 1e-1},
                           {'name': 'x_2', 'initial_value': 2.5, 'bounds': [-5, 5], 'precision': 1e-1},
                           {'name': 'x_3', 'initial_value': 2.5, 'bounds': [-5, 5], 'precision': 1e-1}]
        self.costs = [{'name': 'F'}]

    def evaluate(self, individual):
        x = individual.vector
        result = x[0]**3 + x[1]**2 + x[2]**2
        return [result]


def analytic_gradient(individual):
    x = individual.vector
    dx1 = 3 * x[0]**2
    dx2 = 2 * x[1]
    dx3 = 2 * x[2]
    return [dx1, dx2, dx3]


class TestGradients(unittest.TestCase):
    """ Tests simple one objective optimization problem."""

    def test_local_problem(self):
        problem = ParabolicProblem()
        algorithm = GradientAlgorithm(problem)
        algorithm.options['max_population_size'] = 10
        algorithm.run()

        for individual in problem.individuals:
            gradient = individual.features['gradient']
            ref_gradient = analytic_gradient(individual)
            for i in range(len(gradient)):
                self.assertAlmostEqual(gradient[i], ref_gradient[i], 2)


if __name__ == '__main__':
    unittest.main()



import unittest

from ..benchmark_functions import AlpineFunction, Ackley
from ..benchmark_pareto import ZDT1
from ..algorithm_genetic import NSGAII
from ..results import Results
from ..quality_indicator import epsilon_add


class TestNSGA2(unittest.TestCase):
    def test_local_problem(self):
        problem = ZDT1()
        algorithm = NSGAII(problem)
        algorithm.options['max_population_number'] = 5
        algorithm.options['max_population_size'] = 3
        algorithm.options['max_processes'] = 1
        algorithm.run()

        populations = problem.populations()

        self.assertEqual(len(populations), algorithm.options['max_population_number'] + 1)
        for individuals in populations.values():
            self.assertEqual(len(individuals), algorithm.options['max_population_size'])


class TestZDT1(unittest.TestCase):
    # integration test -- tests the total functionality of nsga2
    # around 11secs according to literature DOI: 10.1007/978-3-642-01020-0_39
    def test_local_problem(self):
        problem = ZDT1()
        algorithm = NSGAII(problem)
        algorithm.options['max_population_number'] = 250
        algorithm.options['max_population_size'] = 100    # according to the literature
        algorithm.options['max_processes'] = 1
        algorithm.run()

        results = Results(problem)
        vals = results.pareto_values()
        exact = problem.pareto_front(vals[0])
        self.assertLessEqual(epsilon_add(exact, vals), 0.2)


class TestAckley(unittest.TestCase):
    """ Tests that the nsga - ii can find the global optimum. """

    def test_local_problem(self, population_number=50):
        try:
            problem = Ackley(**{'dimension': 1})
            algorithm = NSGAII(problem)
            algorithm.options['max_population_number'] = population_number
            algorithm.options['max_population_size'] = 100
            algorithm.options['max_processes'] = 1
            algorithm.run()

            b = Results(problem)
            optimum = b.find_optimum('F_1')  # Takes last cost function
            self.assertAlmostEqual(optimum.costs[0], problem.global_optimum, 1)
        except AssertionError:
            # stochastic
            print("TestAckleyN222::test_local_problem", population_number)
            self.test_local_problem(int(1.5 * population_number))


class TestAlpine(unittest.TestCase):
    """ Tests that the nsga - ii can find the global optimum. """

    def test_local_problem(self, population_number=50):
        try:
            problem = AlpineFunction(**{'dimension': 3})
            algorithm = NSGAII(problem)
            algorithm.options['max_population_number'] = population_number
            algorithm.options['max_population_size'] = 100
            algorithm.options['max_processes'] = 1
            algorithm.run()

            b = Results(problem)
            optimum = b.find_optimum('F_1')  # Takes last cost function
            self.assertAlmostEqual(optimum.costs[0], problem.global_optimum, 1)
        except AssertionError:
            # stochastic
            print("TestAlpine::test_local_problem", population_number)
            self.test_local_problem(int(1.5 * population_number))


if __name__ == '__main__':
    unittest.main()



from .algorithm_genetic import GeneralEvolutionaryAlgorithm
import time
import random


class SweepAlgorithm(GeneralEvolutionaryAlgorithm):
    """
    Sweep Analysis
    """

    def __init__(self, problem, generator, name='SweepAlgorithm'):
        super().__init__(problem, name)
        self.generator = generator

    def run(self):
        t_s = time.time()

        # create initial population
        individuals = self.generator.generate()

        # append to problem
        for individual in individuals:
            self.problem.individuals.append(individual)

        # evaluate individuals
        self.evaluate(individuals)

        t = time.time() - t_s
        self.problem.logger.info("Sweep: elapsed time: {} s".format(t))

        # sync changed individual informations
        self.problem.data_store.sync_all()



# import unittest
# import os
#
# from ..executor import LocalComsolExecutor
# from ..problem import Problem
# from ..algorithm import DummyAlgorithm
# from ..individual import Individual
# from .tests_root import tests_root_path
# __comsol__ = True
# result = os.system('comsol --version')
# if result != 0:
#     __comsol__ = False
#
#
# class ComsolProblem(Problem):
#     """ Describe simple one objective optimization problem. """
#
#     def set(self):
#         self.name = "ComsolProblem"
#         self.parameters = [{'name': 'a', 'initial_value': 10},
#                            {'name': 'b', 'initial_value': 10}]
#         self.costs = [{'name': 'F1', 'criteria': 'minimize'}]
#         self.output_files = ["out.txt"]
#         problem_file = os.path.join(tests_root_path, './data/elstat.mph')
#         self.executor = LocalComsolExecutor(self,
#                                             problem_file=problem_file,
#                                             output_files=self.output_files)
#
#     def evaluate(self, individual):
#         individual.dep_param = 0
#         return self.executor.eval(individual)
#
#     def parse_results(self, output_files, individual):
#         output_file = output_files[0]
#         path = output_file
#         content = ""
#         if os.path.exists(path):
#             with open(path) as file:
#                 content = file.read()
#         else:
#             self.logger.error(
#                 "Output file '{}' doesn't exists.".format(path))
#
#         lines = content.split("\n")
#         line_with_results = lines[5]  # 5th line contains results
#         result = float(line_with_results)
#         return [result]
#
#
# class TestLocalComsol(unittest.TestCase):
#     @unittest.skipIf(__comsol__ is False, "require Comsol Multiphysics")
#     def test_comsol_exec(self):
#         """ Tests one calculation of goal function."""
#         problem = ComsolProblem()
#
#         individuals = [Individual([10, 10])]
#         algorithm = DummyAlgorithm(problem)
#         algorithm.evaluator.evaluate(individuals)
#         try:
#             algorithm.evaluate(individuals)
#             self.assertAlmostEqual(112.94090668383139, individuals[0].costs[0])
#         except Exception as e:
#             print("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")
#             print(e)



import time
import sys
from abc import ABCMeta
from .individual import Individual
from .utils import VectorAndNumbers
from math import inf


class Job(metaclass=ABCMeta):
    def __init__(self, problem):
        self.problem = problem

    def evaluate(self, individual):
        # try to calculate the goal function of the individual in case of failure the individual is
        # replaced by another one, at maximum 5 tries

        # Skips calculation of already calculated individual
        if individual.state == individual.State.EVALUATED:
            return

        for i in range(5):
            # info
            individual.features["start_time"] = time.time()
            t_s = time.time()

            # set in progress
            individual.state = individual.State.IN_PROGRESS

            # check the constraints
            constraints = self.problem.evaluate_constraints(individual)

            if constraints:
                individual.features["feasible"] = sum(map(abs, constraints))

            # problem cost function evaluate only in that case when the problem fits the constraints
            try:
                costs = self.problem.surrogate.evaluate(individual)
                individual.costs = costs
                if self.problem is not None:
                    individual.calc_signed_costs(self.problem.signs)  # the idea is to make this conversion only once

                # set evaluated
                individual.state = individual.State.EVALUATED
                # info
                individual.features["finish_time"] = time.time()
                # write to store
                self.problem.data_store.sync_individual(individual)
                return
            except (TimeoutError, RuntimeError) as e:
                print("Job: error:", e)
                failed_individual = Individual(individual.vector)
                failed_individual.state = individual.State.FAILED
                if individual.features["feasible"] == 0.0:
                    individual.features["feasible"] = inf  # TODO: genetic algorithms uses this information, i dont know the
                    # correct solution
                self.problem.failed.append(failed_individual)
                # in the case of failure generate new random individual
                # TODO: create different strategies
                individual.vector = VectorAndNumbers.gen_vector(self.problem.parameters)
                individual.state = individual.State.EMPTY
                continue
            except:
                print("Job: unexpected error:", sys.exc_info()[0])
                raise

        raise RuntimeError("To many failures has appeared.")



import unittest
from ..algorithm_swarm import OMOPSO, SMPSO
from ..results import Results
from ..benchmark_functions import BinhAndKorn, AlpineFunction, Ackley
from ..benchmark_pareto import ZDT1
from ..quality_indicator import epsilon_add


class TestAckleyOMOPSO(unittest.TestCase):
    """ Tests that the OMOPSO can find the global optimum. """

    def test_local_problem(self, population_number=50):
        problem = Ackley(**{'dimension': 1})
        algorithm = OMOPSO(problem)
        algorithm.options['max_population_number'] = population_number
        algorithm.options['max_population_size'] = 100
        algorithm.options['max_processes'] = 10
        algorithm.options['epsilons'] = 0.1
        algorithm.run()

        b = Results(problem)
        optimum = b.find_optimum('F_1')  # Takes last cost function
        print(optimum.costs[0], problem.global_optimum)
        self.assertAlmostEqual(optimum.costs[0], problem.global_optimum, 1)


class TestZDT1OMOPSO(unittest.TestCase):
    # integration test -- tests the total functionality of OMOPSO

    def test_local_problem(self):
        problem = problem = ZDT1()
        algorithm = OMOPSO(problem)
        algorithm.options['max_population_number'] = 500
        algorithm.options['max_population_size'] = 100  # according to the literature
        algorithm.options['max_processes'] = 1
        algorithm.options['epsilons'] = 0.05
        algorithm.run()

        results = Results(problem)
        vals = results.pareto_values(algorithm.archive)
        exact = problem.pareto_front(vals[0])
        self.assertLessEqual(epsilon_add(exact, vals), 0.2)


class TestAckleySMPSO(unittest.TestCase):
    """ Tests that the SMPSO can find the global optimum. """

    def test_local_problem(self, population_number=50):
        problem = Ackley(**{'dimension': 1})
        algorithm = SMPSO(problem)
        algorithm.options['max_population_number'] = population_number
        algorithm.options['max_population_size'] = 100
        algorithm.options['max_processes'] = 10
        algorithm.run()

        b = Results(problem)
        optimum = b.find_optimum('F_1')  # Takes last cost function
        print(optimum.costs[0], problem.global_optimum)
        self.assertAlmostEqual(optimum.costs[0], problem.global_optimum, 1)


class TestZDT1SMPSP(unittest.TestCase):
    # integration test -- tests the total functionality of SMPSO

    def test_local_problem(self):
        problem = problem = ZDT1()
        algorithm = OMOPSO(problem)
        algorithm.options['max_population_number'] = 500
        algorithm.options['max_population_size'] = 100  # according to the literature
        algorithm.options['max_processes'] = 1
        algorithm.run()

        results = Results(problem)
        vals = results.pareto_values()
        exact = problem.pareto_front(vals[0])
        self.assertLessEqual(epsilon_add(exact, vals), 0.2)


if __name__ == '__main__':
    unittest.main()



import textwrap
import re
import tempfile
import os
import subprocess
import datetime
import time
import ntpath
import pathlib
from string import Template
from uuid import uuid1
from sys import platform

import rpyc
from rpyc.core.async_ import AsyncResultTimeout
from rpyc.utils.classic import upload_file, download_file

from abc import ABCMeta, abstractmethod
from .utils import ConfigDictionary
from shutil import copyfile

from .config import config


# parse ip address
def parse_address(address):
    regex_match = re.compile(r'<(\d{1,3}.\d{1,3}.\d{1,3}.\d{1,3}):.*').match(address)
    if regex_match is not None:
        return regex_match.group(1)
    return ""


class Executor(metaclass=ABCMeta):
    """
    Function is a class representing objective or cost function for
    optimization problems.
    """

    def __init__(self, problem):
        self.uuid = uuid1().hex
        self.problem = problem

        self.options = ConfigDictionary()

        # set default parse method from problem (can be overridden)
        if "parse_results" in dir(self.problem):
            self.parse_results = self.problem.parse_results

    @abstractmethod
    def eval(self, individual):
        pass

    @staticmethod
    def _join_parameters_names(parameters, sep=","):
        param_names_string = ""

        for parameter in parameters:
            param_names_string += str(parameter['name']) + sep

        # remove last sep
        if (len(parameters)) > len(sep):
            param_names_string = param_names_string[:-len(sep)]

        return param_names_string

    @staticmethod
    def _join_parameters_values(x, sep=","):
        param_values_string = ""
        for number in x:
            param_values_string += str(number) + sep

        # remove last sep
        if (len(x)) > len(sep):
            param_values_string = param_values_string[:-len(sep)]

        return param_values_string

    @staticmethod
    def _remove_dir(local_dir):
        files = os.listdir(path=local_dir)
        for f in files:
            filepath = os.path.join(local_dir, f)
            os.remove(filepath)

        os.rmdir(local_dir)


class LocalComsolExecutor(Executor):
    comsol_command = Template("comsol batch -inputfile $input_file -nosave -pname $param_names -plist $param_values")

    def __init__(self, problem, problem_file, input_files=None, output_files=None):
        super().__init__(problem)
        self.model_file = ntpath.basename(problem_file)
        copyfile(problem_file, self.problem.working_dir + self.model_file)

        self.output_files = output_files
        self.input_files = input_files

    def eval(self, individual):
        super().eval(individual)

        param_names_string = Executor._join_parameters_names(self.problem.parameters)
        param_values_string = Executor._join_parameters_values(individual.vector)

        cmd_string = self.comsol_command.substitute(input_file=self.model_file,
                                                    param_names=param_names_string,
                                                    param_values=param_values_string)

        try:
            # run command
            current_path = os.getcwd()
            os.chdir(self.problem.working_dir)

            out = subprocess.run(cmd_string, shell=True)
            os.chdir(current_path)

            if out.returncode != 0:
                err = "Unknown error"
                if out.stderr is not None:
                    err = "Cannot run COMSOL Multiphysics.\n\n {}".format(out.stderr)

                self.problem.logger.error(err)
                raise RuntimeError(err)

            output_files = []
            for file in self.output_files:
                output_files.append(self.problem.working_dir + file)
            result = self.parse_results(output_files, individual)

            return result
        except Exception as e:
            err = "Cannot run COMSOL Multiphysics.\n\n {}".format(e)
            self.problem.logger.error(err)
            raise RuntimeError(err)


class LocalFEMMExecutor(Executor):
    """
    Appends the problem parameters to the beginning of the script file, than runs the FEMM in a subprocess and retrieves
    the data in this command window.
    """

    def __init__(self, problem, script_file, output_files):
        super().__init__(problem)
        self.script_file = ntpath.basename(script_file)

        self.femm_command = 'wine ~/.wine/drive_c/femm42/bin/femm.exe'

        self.output_files = output_files

    def eval(self, individual):
        super().eval(individual)

        param_names_string = Executor._join_parameters_names(self.problem.parameters)
        param_values_string = Executor._join_parameters_values(individual.vector)

        lua_path = os.path.abspath(self.script_file)
        if os.path.isfile(lua_path) and platform == 'linux':
            arg = '"' + os.popen('winepath -w "' + lua_path + '"').read().strip() + '"'

        # cmd_string = self.femm_command + ' -lua-script={} -lua-var={} -lua-var={} -windowhide'.format(arg, 'radius=200',
        #                                                                                              'c=42')
        cmd_string = self.femm_command + ' -lua-script={}'.format(arg)

        params = param_names_string.split(',')
        values = param_values_string.split(',')
        #
        for i in range(len(params)):
            temp = str(params[i])+'='+str(values[i])
            cmd_string += ' -lua-var={}'.format(temp)

        try:

            out = subprocess.run(cmd_string, shell=True, stdout=subprocess.PIPE)

            if out.returncode != 0:
                err = "Unknown error"
                if out.stderr is not None:
                    err = "Cannot run FEMM.\n\n {}".format(out.stderr)

                self.problem.logger.error(err)
                raise RuntimeError(err)

            result = self.parse_results(self.output_files, individual)
            return result

        except Exception as e:
            err = "Cannot run FEMM with wine.\n\n {}".format(e)
            self.problem.logger.error(err)
            raise RuntimeError(err)



class LocalAnsysExecutor(Executor):
    """
    runs an ansys executable python script
    """

    def __init__(self, problem, script_file, output_files):
        super().__init__(problem)
        self.script_file = ntpath.basename(script_file)

        self.ansys_path = r'C:\Program Files\AnsysEM\AnsysEM20.2\Win64\ansysedt.exe'
        #"D:\tmp\Example_Conductor.py"
        self.output_files = output_files

    def eval(self, individual):
        super().eval(individual)

        param_names_string = Executor._join_parameters_names(self.problem.parameters)
        param_values_string = Executor._join_parameters_values(individual.vector)

        #lua_path = os.path.abspath(self.script_file)
        #if os.path.isfile(lua_path) and platform == 'linux':
        #    arg = '"' + os.popen('winepath -w "' + lua_path + '"').read().strip() + '"'

        cmd_string = self.ansys_path + ' - runscript'.format(self.script_file)

        try:

            out = subprocess.run(cmd_string, shell=True, stdout=subprocess.PIPE)

            if out.returncode != 0:
                err = "Unknown error"
                if out.stderr is not None:
                    err = "Cannot run Ansys.\n\n {}".format(out.stderr)

                self.problem.logger.error(err)
                raise RuntimeError(err)

            result = self.parse_results(self.output_files, individual)
            return result

        except Exception as e:
            err = "Cannot run Ansys.\n\n {}".format(e)
            self.problem.logger.error(err)
            raise RuntimeError(err)


class RemoteExecutor(Executor):
    """
        Allows distributing of calculation of objective functions.
        """

    def __init__(self, problem, command, files_to_server, files_from_server=None):
        super().__init__(problem)

        if problem.working_dir is None:
            raise Exception('RemoteExecutor: Problem working directory must be set.')

        self.options.declare(name='hostname', default=config["condor_host"],
                             desc='Hostname')
        self.options.declare(name='username', default=config["condor_login"],
                             desc='Username')
        self.options.declare(name='port', default=15900, lower=0,
                             desc='Port')

        # command
        self.command = command

        # files
        self.input_files = files_to_server
        self.output_files = files_from_server

    @abstractmethod
    def eval(self, individual):
        super().eval(individual)

    def _create_client(self):
        # key = os.path.join(os.path.dirname(__file__), "cert/artap.key")
        # cert = os.path.join(os.path.dirname(__file__), "cert/artap.crt")

        channel = rpyc.Channel(rpyc.SocketStream.connect(self.options["hostname"], self.options["port"]))
        channel.send(self.options["username"].encode('utf-8'))
        response = channel.recv()
        AUTH_ERROR = b'error'
        if response == AUTH_ERROR:
            raise rpyc.utils.authenticators.AuthenticationError('Invalid username for daemon')

        return rpyc.utils.factory.connect_channel(channel, service=rpyc.core.ClassicService)
        # return rpyc.classic.connect(self.options["hostname"], self.options["port"])
        # return rpyc.classic.ssl_connect(self.options["hostname"], self.options["port"], keyfile=key, certfile=cert)

    def _init_remote(self, client):
        try:
            # create client
            remote_dir = client.root.create_job_dir()
            return remote_dir
        except RuntimeError as e:
            # self.problem.logger.error("Cannot create remote directory '{}' on ''".format(directory,
            #                                                                             self.options["hostname"]))
            print("ERROR --------------------------------")
            pass

    @staticmethod
    def _create_file_on_remote(destination_file, content, remote_dir, client):
        fp = tempfile.NamedTemporaryFile(mode='w+t', delete=False, newline="\n")
        fp.write(content)
        fp.close()

        RemoteExecutor._transfer_file_to_remote(fp.name, destination_file, remote_dir, client)

        os.unlink(fp.name)

    @staticmethod
    def _transfer_file_to_remote(source_file, destination_file, remote_dir, client):
        source = source_file
        dest = "{}/{}".format(remote_dir, destination_file)

        upload_file(client, localpath=source, remotepath="{}/{}".format(client.root.artap_dir, dest))

    @staticmethod
    def _transfer_file_from_remote(source_file, destination_file, remote_dir, client):
        dest = destination_file
        source = "{}/{}".format(remote_dir, source_file)

        download_file(client, remotepath="{}/{}".format(client.root.artap_dir, source), localpath=dest)

    @staticmethod
    def _create_dir_on_remote(directory, client):
        d = datetime.datetime.now()
        ts = d.strftime("%Y-%m-%d-%H-%M-%S-%f")

        # projects directory
        client.exec_command("mkdir " + directory)

        # working directory
        wd = "{}/artap-{}".format(directory, ts)
        client.exec_command("mkdir " + wd)

        return wd

    def _transfer_files_to_remote(self, remote_dir, client):
        # transfer input files
        if self.input_files:
            for file in self.input_files:
                self._transfer_file_to_remote("{}/{}".format(self.problem.working_dir, file), "./{}".format(file),
                                              remote_dir=remote_dir, client=client)

    def _transfer_files_from_remote(self, client):
        pass


class CondorJobExecutor(RemoteExecutor):
    """
    Allows distributing of calculation of objective functions.
    """

    def __init__(self, problem, files_to_condor, files_from_condor=None):
        super().__init__(problem, "", files_to_condor, files_from_condor)

        # set default host
        self.options["hostname"] = config["condor_host"]

        # set default requirements
        self.requirements = ""
        self.request_cpus = -1
        self.request_memory = -1
        self.hold_on_start = False

        # executor id -  create client
        # check if running on condor
        if os.path.exists('.job.ad'):
            # read condor job ad
            with open(".job.ad", 'r') as file:
                for line in file.readlines():
                    # update remote dir with executor_id
                    if line.startswith("OrigIwd = "):
                        remote_dir = line[11:-2].split("/")[-1]
                        client = self._create_client()
                        client.root.log_update_executor(remote_dir, self.uuid)

    @abstractmethod
    def _create_job_file(self, remote_dir, individual, client):
        pass

    def _create_desc(self, individual):
        desc = {}
        desc["executor_id"] = self.uuid
        desc["individual_id"] = individual.id
        desc["individual_population_id"] = individual.population_id
        desc["individual_algorithm_id"] = individual.algorithm_id
        desc["individual_vector"] = '[' + ', '.join(str(e) for e in individual.vector) + ']'

        return desc

    def eval(self, individual):
        if self.options["hostname"] is None:
            raise Exception("Condor host is not defined.")

        super().eval(individual)

        success = False
        while not success:
            try:
                # create client
                client = self._create_client()

                # init remote
                remote_dir = self._init_remote(client=client)
                # self.problem.logger.info("RemoteDir {}".format(remote_dir))

                # transfer supplementary files, input and model file
                self._transfer_files_to_remote(remote_dir, client)

                # submit job
                try:
                    self._create_job_file(remote_dir, individual, client)
                except AsyncResultTimeout as e:
                    print("ERROR - {} - try again ({})".format(e, remote_dir))
                    # remove job dir
                    client.root.remove_job_dir(remote_dir)
                    client.close()
                    # resubmit
                    return self.eval(individual)

                start = time.time()

                events = []
                cnt = 0
                successful_job = False
                delay = 0.5

                run = True
                while run:
                    eventlog = client.root.eventlog(remote_dir)
                    for e in eventlog:
                        if {e["timestamp"], e["type"]} not in events:
                            events.append({e["timestamp"], e["type"]})

                    # print("len(events) = {}".format(len(events)))
                    if len(events) > 0:
                        for i in range(cnt, len(events)):
                            event = eventlog[i]
                            tp = event["type"]

                            args = ""

                            if tp == "submit":
                                # SUBMIT
                                pass
                            elif tp == "execute":
                                # EXECUTE
                                args += "ExecuteHost: {}, ".format(event["execute_host"])
                                pass
                            elif tp == "image_size":
                                # IMAGE_SIZE
                                pass
                            elif tp == "job_terminated":
                                # JOB_TERMINATED
                                successful_job = event["successful"]
                                run = False
                            elif tp == "job_held":
                                # JOB_HELD

                                self.problem.logger.error(
                                    "Job {}.{} is '{}' at {}".format(event["cluster"], event["proc"], event["type"],
                                                                     ""))
                                run = False
                                # read log
                                # content_log = self._read_file_from_remote("{}.log".format(self.output_files[0]),
                                #                                          remote_dir=remote_dir, client=client)
                                # self.problem.logger.error(content_log)
                                # remove job
                                # self._run_command_on_remote("condor_rm {}".format(process_id),
                                #                            remote_dir=remote_dir, client=client)
                                raise RuntimeError

                            if len(args) > 0:
                                args = args[:-2]

                            self.problem.logger.info(
                                "Job {}.{} ({}) is '{}' at {}".format(event["cluster"], event["proc"], remote_dir,
                                                                      event["type"], args))
                            # print("{}: {} ({})".format(eventlog[i].timestamp, eventlog[i].type, args))

                    if run:
                        cnt = len(events)
                        time.sleep(delay)

                end = time.time()
                if (end - start) > self.problem.options["time_out"]:
                    raise TimeoutError

                if successful_job:
                    if len(self.output_files) > 0:
                        output_files = []
                        d = datetime.datetime.now()
                        ts = d.strftime("%Y-%m-%d-%H-%M-%S-%f")
                        path = self.problem.working_dir + 'artap' + ts
                        os.mkdir(path)

                        for file in self.output_files:
                            self._transfer_file_from_remote(source_file=file,
                                                            destination_file="{}/{}".format(path, file),
                                                            remote_dir=remote_dir, client=client)
                            output_files.append("{}/{}".format(path, file))
                        success = True
                        result = self.parse_results(output_files, individual)
                        # update cost on remote server
                        client.root.log_update_cost(remote_dir, individual, result)

                    # remove job dir
                    if result is not None:
                        client.root.remove_job_dir(remote_dir)

                    if self.problem.options['save_data_files'] is False:
                        self._remove_dir(path)
                else:
                    assert 0

                # remove job dir
                # client.root.remove_job_dir(remote_dir)

                client.close()

                return result

            except ConnectionError as e:
                print(e)
                time.sleep(1.0)
                continue


class CondorPythonJobExecutor(CondorJobExecutor):
    def __init__(self, problem, script, parameter_file, output_files=None, python_path="python3"):
        self.script = ntpath.basename(script)
        super().__init__(problem, [self.script], output_files)

        self.parameter_file = parameter_file
        copyfile(script, self.problem.working_dir + self.script)

        self.executable = textwrap.dedent("""\
            #!/bin/sh
            # args

            """ + python_path + """  $@""")

        self.requirements = "(OpSys == \"LINUX\" && Arch == \"X86_64\")"

    def _create_job_file(self, remote_dir, individual, client):
        condor_output_files = self.output_files

        condor_input_files = [self.script]
        # create input file with parameters

        if self.parameter_file:
            # parameters
            # create remote file
            param_values_string = Executor._join_parameters_values(individual.vector, "\n")
            condor_input_files.append(self.parameter_file)
            arguments = self.script + " " + self.parameter_file
            self._create_file_on_remote(self.parameter_file, param_values_string, remote_dir=remote_dir,
                                        client=client)
        else:
            param_values_string = Executor._join_parameters_values(individual.vector, ",")
            arguments = self.script + " " + param_values_string

        # create executable
        self._create_file_on_remote("run.sh", self.executable, remote_dir=remote_dir, client=client)

        # desc
        desc = self._create_desc(individual)
        desc["type"] = "python"
        desc["extension"] = ".py"
        desc["editor"] = True
        desc["name"] = "Python"

        client.root.submit_job(remote_dir=remote_dir,
                               executable="{}/{}/run.sh".format(client.root.artap_dir, remote_dir),
                               arguments=arguments,
                               input_files=condor_input_files,
                               output_files=condor_output_files,
                               requirements=self.requirements,
                               request_cpus=self.request_cpus,
                               request_memory=self.request_memory,
                               hold_on_start=self.hold_on_start,
                               desc=desc)


class CondorMatlabJobExecutor(CondorJobExecutor):
    def __init__(self, problem, script, parameter_file, files_from_condor=None,
                 matlab_path="/opt/matlab-R2018b/bin/matlab"):
        self.script = ntpath.basename(script)
        super().__init__(problem, [self.script], files_from_condor)
        self.parameter_file = parameter_file
        copyfile(script, self.problem.working_dir + self.script)

        self.executable = textwrap.dedent("""\
            #!/bin/sh
            # args
            s=$@

            """ + matlab_path + """ -nodisplay -nosplash -nodesktop -r ${s%.m}""")

        self.requirements = "(OpSys == \"LINUX\" && Arch == \"X86_64\")"

    def _create_job_file(self, remote_dir, individual, client):
        condor_input_files = [self.parameter_file, self.input_files[0]]
        condor_output_files = self.output_files

        # create input file with parameters
        if self.parameter_file:
            # parameters
            param_values_string = Executor._join_parameters_values(individual.vector, "\n")
            # create remote file
            self._create_file_on_remote(self.parameter_file, param_values_string, remote_dir=remote_dir, client=client)

        # create executable
        self._create_file_on_remote("run.sh", self.executable, remote_dir=remote_dir, client=client)

        # desc
        desc = self._create_desc(individual)
        desc["type"] = "matlab"
        desc["extension"] = ".m"
        desc["editor"] = False
        desc["name"] = "Matlab"

        client.root.submit_job(remote_dir=remote_dir,
                               executable="{}/{}/run.sh".format(client.root.artap_dir, remote_dir),
                               arguments=self.script,
                               input_files=condor_input_files,
                               output_files=condor_output_files,
                               requirements=self.requirements,
                               request_cpus=self.request_cpus,
                               request_memory=self.request_memory,
                               hold_on_start=self.hold_on_start,
                               desc=desc)


class CondorComsolJobExecutor(CondorJobExecutor):
    def __init__(self, problem, model_file, files_from_condor=None, comsol_path="/opt/comsol-5.4/bin/comsol"):
        self.model_file = ntpath.basename(model_file)
        super().__init__(problem, [self.model_file], files_from_condor)
        copyfile(model_file, self.problem.working_dir + self.model_file)

        self.arguments = Template("-inputfile $input_file -nosave -pname $param_names -plist $param_values")
        self.executable = textwrap.dedent("""\
                #!/bin/sh
                """ + comsol_path + """ batch $@
                """)

        self.request_cpus = 3
        self.request_memory = 30
        self.requirements = "(OpSys == \"LINUX\" && Arch == \"X86_64\")"

    def _create_job_file(self, remote_dir, individual, client):
        condor_input_files = [self.model_file]
        condor_output_files = self.output_files

        param_names_string = Executor._join_parameters_names(self.problem.parameters)
        param_values_string = Executor._join_parameters_values(individual.vector)
        arguments = self.arguments.substitute(input_file=os.path.basename(self.model_file),
                                              param_names=param_names_string,
                                              param_values=param_values_string)

        # create executable
        self._create_file_on_remote("run.sh", self.executable, remote_dir=remote_dir, client=client)

        # desc
        desc = self._create_desc(individual)
        desc["type"] = "comsol"
        desc["extension"] = ".mph"
        desc["editor"] = False
        desc["name"] = "Comsol Multiphysics"

        client.root.submit_job(remote_dir=remote_dir,
                               executable="{}/{}/run.sh".format(client.root.artap_dir, remote_dir),
                               arguments=arguments,
                               input_files=condor_input_files,
                               output_files=condor_output_files,
                               requirements=self.requirements,
                               request_cpus=self.request_cpus,
                               request_memory=self.request_memory,
                               hold_on_start=self.hold_on_start,
                               desc=desc)


class CondorCSTJobExecutor(CondorJobExecutor):
    def __init__(self, problem, model_file, files_from_condor=None,
                 cst_path="\"C:\Program Files (x86)\CST Studio Suite 2020\CST DESIGN ENVIRONMENT.exe\""):
        self.model_file = ntpath.basename(model_file)
        super().__init__(problem, [self.model_file], files_from_condor)
        copyfile(model_file, self.problem.working_dir + self.model_file)

        self.executable = textwrap.dedent("""\
            """ + cst_path + """ --se --rebuild --hide -par parameters.txt -project-file {} """.format(self.model_file))
        self.executable += '\n'
        self.executable += '"C:\Program Files\\7-Zip\\7z.exe" a "{}.zip" "{}\\"'.format(
            os.path.splitext(self.model_file)[0], os.path.splitext(self.model_file)[0])
        self.executable += '\n'

        i = 1
        for file in files_from_condor[1:]:
            filename = pathlib.Path(file)
            windows_path = pathlib.PureWindowsPath(filename)
            name = filename.name
            self.output_files[i] = name
            self.executable += 'copy "{}" "{}" \n'.format(windows_path, name)
            i += 1

        # add solver directory to output
        # self.output_files.append(os.path.splitext(self.model_file)[0] + "\Export")

        self.request_cpus = 2
        self.request_memory = 10
        self.requirements = "(OpSys == \"WINDOWS\" && Arch == \"X86_64\")"

    def _create_job_file(self, remote_dir, individual, client):
        condor_input_files = [self.model_file]
        condor_output_files = self.output_files

        parameters = ""
        for parameter, value in zip(self.problem.parameters, individual.vector):
            parameters += "{}={}\n".format(parameter['name'], value)

        parameter_file = "parameters.txt"
        condor_input_files.append(parameter_file)
        self._create_file_on_remote(parameter_file, parameters, remote_dir=remote_dir,
                                    client=client)

        # create executable
        self._create_file_on_remote("run.bat", self.executable, remote_dir=remote_dir, client=client)

        # desc
        desc = self._create_desc(individual)
        desc["type"] = "cst"
        desc["extension"] = ".cst"
        desc["editor"] = False
        desc["name"] = "CST"

        client.root.submit_job(remote_dir=remote_dir,
                               executable="{}/{}/run.bat".format(client.root.artap_dir, remote_dir),
                               arguments="",
                               input_files=condor_input_files,
                               output_files=condor_output_files,
                               requirements=self.requirements,
                               request_cpus=self.request_cpus,
                               request_memory=self.request_memory,
                               hold_on_start=self.hold_on_start,
                               desc=desc)



"""
algorithm.py
===========================================
This is base class for all algorithms
"""

from .problem import Problem
from .utils import ConfigDictionary
from abc import ABCMeta
from .operators import Evaluator, GradientEvaluator, WorstCaseEvaluator

from enum import Enum
from uuid import uuid1


class EvaluatorType(Enum):
    SIMPLE = 0
    GRADIENT = 1
    WORST_CASE = 2


class Algorithm(metaclass=ABCMeta):
    """ Base class for optimization algorithms. """

    def __init__(self, problem: Problem, name="Algorithm", evaluator_type=EvaluatorType.SIMPLE):
        self.uuid = uuid1().hex
        self.name = name
        self.problem = problem
        if evaluator_type == EvaluatorType.SIMPLE or evaluator_type is None:
            self.evaluator = Evaluator(self)
        elif evaluator_type == EvaluatorType.GRADIENT:
            self.evaluator = GradientEvaluator(self)
        elif evaluator_type == EvaluatorType.WORST_CASE:
            self.evaluator = WorstCaseEvaluator(self)

        self.parameters = problem.parameters
        self.options = ConfigDictionary()

        self.options.declare(name='verbose_level', default=1, lower=0,
                             desc='Verbose level')
        self.options.declare(name='calculate_gradients', default=False,
                             desc='Enable calculating of gradients')
        # max(int(2 / 3 * cpu_count(), 1)
        self.options.declare(name='max_processes', default=1,
                             desc='Max running processes')

        self.options.declare(name='n_iterations', default=10,
                             desc='Max number of iterations')

        self.individual_features = dict()

    def evaluate(self, individuals):
        # set algorithm id
        for individual in individuals:
            individual.algorithm_id = self.uuid

        self.evaluator.evaluate(individuals)

    def evaluate_scalar(self, individual):
        # set algorithm id
        individual.algorithm_id = self.uuid

        self.evaluator.evaluate_scalar(individual)
        return individual.costs[0]


class DummyAlgorithm(Algorithm):
    """
    Dummy class for testing
    """

    def __init__(self, problem, name='DummyAlgorithm'):
        super().__init__(problem, name)

    def run(self):
        pass




from .surrogate import SurrogateModelPredict

import numpy as np

from smt.surrogate_models import RBF


class SurrogateModelSMT(SurrogateModelPredict):
    def __init__(self, problem):
        super().__init__(problem)

        self.train_step = -1
        self.sigma_threshold = 10
        self.score_threshold = 0.5

    def init_default_regressor(self):
        # default regressor
        self.regressor = RBF(d0=5, print_prediction=False)
        self.has_epsilon = False

    def predict(self, x, *args):
        if self.trained:
            return self.regressor.predict_values(np.array([x]))
        else:
            assert 0

    def predict_variances(self, x, *args):
        if self.trained:
            return self.regressor.predict_variances(np.array([x]), *args)
        else:
            assert 0

    def train(self):
        self.trained = False
        assert(len(self.x_data) == len(self.y_data))

        # print(self.x_data)
        # print("Trained set: {}".format(len(self.x_data)))

        self.regressor.options["print_global"] = False
        self.regressor.set_training_values(np.array(self.x_data), np.array(self.y_data))
        self.regressor.train()

        if self.eval_stats:
            # score
            pass
            # self.score = self.regressor.score(self.x_data, self.y_data)
            # print("self.score = {} : {}".format(len(self.x_data), self.score))
            # lml (Gaussian regressor)
            # if "log_marginal_likelihood" in dir(self.regressor):
            #    self.lml, self.lml_gradient = self.regressor.log_marginal_likelihood(self.regressor.kernel_.theta, eval_gradient=True)

        self.trained = True




from numpy.random import random_sample
from .problem import Problem
from .algorithm_swarm import SwarmAlgorithm
from .operators import DummySelector, RandomGenerator, ParetoDominance, crowding_distance, nondominated_truncate, \
    Operator
from math import exp
from random import random
import time


class MoFirefly(SwarmAlgorithm):
    """
    Multi-objective firefly algorithm based on [4].

    The single-objective firefly algorithm is a modification of the original pso algorithms. The idea is that it mimics
    the behaviour of the fireflies, which uses specfic light combinations for hunting and dating. This algorithm mimics
    the dating behaviour of these bugs. The algorithm is originally published by [1]

    The brightest individual attracts the darkest ones, this starts to go to that direction, if there is not an
    existing brighter solutions, the algorithm randomly steps one into another direction.

    This operator makes the following
    ---------------------------------

    - calculates the distance between two points, because its correlates with that value.
    - the brightness of the other point
    - the mutator constant, which contains a damping factor [2]?

    The k+1 th position of the j(th) individual is calculated by the following formula
    ---------------------------------------------------------------------------

        b0=2;               # Attraction Coefficient Base Value
        a=0.2;              # Mutation Coefficient
        ad=0.98;            # Mutation Coefficient Damping Ratio -- decreases the initial value of a after each
                              iteration step
        gamma = 1.          # light absorbtion coefficient

        x(j,k+1) = x(j,k) + b0*exp(-gamma*r^2) + sum_{i<j}[x(j,k) - x(i,k)] + a*(rand[0,1] - 0.5)

    [1] Yang, Xin-She. "Firefly algorithms for multimodal optimization." International symposium on stochastic algorithms.
        Springer, Berlin, Heidelberg, 2009.
    [2] firefly algorithm implementation from www.Yarpiz.com

    [3] https://nl.mathworks.com/matlabcentral/fileexchange/29693-firefly-algorithm

    [4]: Yang, X. S. (2013). Multiobjective firefly algorithm for continuous optimization.
            Engineering with computers, 29(2), 175-184.

     Similarly, alpha should also be linked with scales, the steps should not too large or too small, often steps
     are about 1/10 to 1/100 of the domain size. In addition, alpha should be reduced gradually using
     alpha=alpha_0 delta^t during eteration t.  Typically, delta=0.9 to 0.99 will be a good choice.
    """

    def __init__(self, problem: Problem, name="MOFA Algorithm"):
        super().__init__(problem, name)
        self.options.declare(name='prob_mutation', default=0.1, lower=0,
                             desc='prob_mutation'),
        self.n = self.options['max_population_size']

        self.selector = DummySelector(self.problem.parameters)
        self.dominance = ParetoDominance()

        # set random generator
        self.individual_features['dominate'] = []
        self.individual_features['crowding_distance'] = 0
        self.individual_features['domination_counter'] = 0
        self.individual_features['front_number'] = 0
        self.generator = RandomGenerator(self.problem.parameters, self.individual_features)

        # default parameters
        self.beta = 2.0  # attraction constant
        self.alpha = 0.2  # mutation coefficient --- the original coefficient for the single problem
        self.ad = 0.98  # mutation coefficient -- damping ratio decreases the initial value of a after each
        self.gamma = 1.  # light absorbtion coefficient

        # self.mutator = PmMutator(self.problem.parameters, self.options['prob_mutation'])
        # constants for the speed and the position calculation

    def turbulence(self, particles, current_step=0):
        """ Lévy flight """

        for i in range(len(particles)):
            if i % 6 == 0:
                mutated = self.mutator.mutate(particles[i])
                particles[i].vector = copy(mutated.vector)

    def step(self, current, other):
        """
        If the other firefly is more attractive the current firefly steps one step into the other fireflies direction.
        :return: the new position of the firefly.
        """

        # if it's dominant returns without changing its position
        if self.dominance.compare(current, other) == 1:
            return
        else:
            for i, param in enumerate(self.parameters):
                lb = param['bounds'][0]
                ub = param['bounds'][1]

                # elementary distance of the particle, random walk
                e = random * (ub - lb)
                # distance between the two individuals
                r2 = current.vector[i] ** 2. + other.vector[i] ** 2.
                vel_attraction = self.beta * exp(-self.gamma * r2 ** 2.) * (other.vector[i] - current.vector[i])

                v_rd = self.alpha * (random() - 0.5) * e
                # position i+1
                current.vector[i] += vel_attraction + v_rd
        return

    # def update_position(self, population):
    #
    #     for individual in population:
    #         for parameter, i in zip(self.parameters, range(len(individual.vector))):
    #             individual.vector[i] = individual.vector[i] + individual.features['velocity'][i]
    #
    #             # adjust maximum position if necessary
    #             if individual.vector[i] > parameter['bounds'][1]:
    #                 individual.vector[i] = parameter['bounds'][1]
    #                 individual.features['velocity'][i] *= 0.001
    #
    #             # adjust minimum position if necessary
    #             if individual.vector[i] < parameter['bounds'][0]:
    #                 individual.vector[i] = parameter['bounds'][0]
    #                 individual.features['velocity'][i] *= 0.001

    def update_global_best(self, offsprings):
        """ Manages the leader class in OMOPSO. """

        # the fitness of the particles are calculated by their crowding distance
        crowding_distance(swarm)

        # the length of the leaders archive cannot be longer than the number of the initial population
        self.leaders += swarm
        self.leaders.truncate(self.options['max_population_size'], 'crowding_distance')
        # self.problem.archive += swarm

        return

    def run(self):
        t_s = time.time()
        self.problem.logger.info("PSO: {}/{}".format(self.options['max_population_number'],
                                                     self.options['max_population_size']))
        # initialize the swarm
        self.generator.init(self.options['max_population_size'])
        individuals = self.generator.generate()
        self.evaluate(individuals)
        self.add_features(individuals)

        self.update_global_best(individuals)

        i = 0
        while i < self.options['max_population_number']:
            offsprings = self.selector.select(individuals)

            self.update_velocity(offsprings)
            self.update_position(offsprings)
            self.turbulence(offsprings, i)

            self.evaluate(offsprings)

            self.update_particle_best(offsprings)
            self.update_global_best(offsprings)

            i += 1

        t = time.time() - t_s
        self.problem.logger.info("PSO: elapsed time: {} s".format(t))

    # def __init__(self, problem: Problem, name="Particle Swarm Algorithm"):
    #     super().__init__(problem, name)
    #     self.n = self.options['max_population_size']
    #     self.mutator = FireflyStep(self.problem.parameters)
    #     self.selector = DummySelector(self.problem.parameters, self.problem.signs)
    #
    # def run(self):
    #     self.generator = RandomGenerator(self.problem.parameters)
    #     self.generator.init(self.options['max_population_size'])
    #
    #     population = self.gen_initial_population()
    #     self.selector.fast_nondominated_sorting(population.individuals)
    #
    #     t_s = time.time()
    #     self.problem.logger.info("PSO: {}/{}".format(self.options['max_population_number'],
    #                                                  self.options['max_population_size']))
    #
    #     nr_gen = 0
    #     non_dominated_sol = []
    #     while nr_gen < self.options['max_population_number']:
    #         offsprings = self.selector.select(population.individuals)
    #         for i, a in enumerate(offsprings):
    #             non_dominated = True
    #             for j, b in enumerate(offsprings):
    #                 if i != j:
    #                     if self.mutator.dominate(a, b):
    #                         # in this case the individual a moves one in the direction b
    #                         # the step size is calculated from the intensity and the attraction
    #                         self.mutator.mutate_ij(a, b)
    #                         non_dominated = False
    #                         # self.evaluate([a]) <- if its not true a new one should be generated, this should be
    #                         #                      managed by the parallelization of the evaluator
    #             if non_dominated:
    #                 non_dominated_sol.append(a) # collects the non-dominated solutions for an iteration
    #
    #         self.evaluate(offsprings)
    #
    #         # to improve the convergence one solution from the pareto front is selected as global best
    #         # generating random weights for the cost function to select the 'global best'
    #         weights = random_sample(len(self.problem.costs))
    #         c = 1. / sum(weights)
    #         weights = [x * c for x in weights]
    #
    #         self.problem.populations[-1] = offsprings
    #
    #         nr_gen += 1
    #
    #     t = time.time() - t_s
    #     self.problem.logger.info("PSO: elapsed time: {} s".format(t))
